#!/bin/bash

# echo commands
set -x
# exit on error
set -e

# save folder of script for location of the patches
patch_folder=$(readlink -f $(dirname $0))

# install necessary system packages
sudo apt-get install --assume-yes git mercurial cmake g++ libboost-all-dev bc flex bison libreadline-dev ocaml ocaml-native-compilers locales
if grep -q 'Ubuntu 16\.04' /etc/lsb-release 2>/dev/null; then
    # newer stack and ghc needed for Ubuntu 16.04
    wget -qO- https://get.haskellstack.org/ | sh
    # need no force overwrite to install openjdk-9-jdk on Ubuntu 16.04 due to broken package dependencies
    sudo apt-get --assume-yes -o Dpkg::Options::="--force-overwrite" install openjdk-9-jdk
    # need newer version of gradle
else
    # install system ghc, stack, and jdk for Ubuntu 17.10
    sudo apt-get install --assume-yes ghc haskell-stack openjdk-9-jdk gradle
fi

# generate local for number formats
sudo locale-gen en_US.UTF-8

# collect binaries in this folder
cd ~
mkdir -p ~/bin
export PATH=$PATH:$HOME/bin
echo "export PATH=\$PATH:\$HOME/bin" >>~/.bashrc

# install combine-aiger (use sparse checkout)
cd ~
mkdir syntcomp
cd syntcomp
git init
git remote add -f origin https://bitbucket.org/swenjacobs/syntcomp.git
git config core.sparsecheckout true
echo "tools/aiger-ltl-model-checker/" >> .git/info/sparse-checkout
git pull origin master
cd tools/aiger-ltl-model-checker
make
cp combine-aiger ~/bin
cd ~
rm -rf syntcomp

# install syfco
cd ~
git clone https://github.com/reactive-systems/syfco.git
cd syfco
if grep -q 'Ubuntu 16\.04' /etc/lsb-release 2>/dev/null; then
    # newer ghc and stack needed for Ubuntu 16.04
    /usr/local/bin/stack setup
    /usr/local/bin/stack build
    cp $(/usr/local/bin/stack path | grep local-install-root | sed 's/local-install-root: //')/bin/syfco ~/bin
else
    make
    cp syfco ~/bin
fi
cd ~
rm -rf syfco

# install abc
hg clone https://bitbucket.org/alanmi/abc
cd abc
make
cp abc ~/bin
rm -rf abc

# install newer ocamlbuild to compile speculoos (only needed for Ubuntu 17.10)
if grep -q 'Ubuntu 17\.10' /etc/lsb-release 2>/dev/null; then
    cd ~
    wget http://archive.ubuntu.com/ubuntu/pool/universe/o/ocamlbuild/ocamlbuild_0.11.0.orig.tar.gz
    tar -xzf ocamlbuild_0.11.0.orig.tar.gz
    cd ocamlbuild-0.11.0
    make
    sudo CHECK_IF_PREINSTALLED=false make install
    cd ~
    rm -r ocamlbuild-0.11.0 ocamlbuild_0.11.0.orig.tar.gz
fi

# install speculoos dependencies
cd ~
git clone https://github.com/romainbrenguier/ocaml-cudd.git
git clone https://github.com/romainbrenguier/ocaml-aiger.git
cd ocaml-cudd
make
cp cudd_stub.mli cudd_stub.ml
ocamlbuild cudd_stub.cmo
cd ~
cd ocaml-aiger
make
cd ~

# install Speculoos
cd ~
git clone https://github.com/romainbrenguier/Speculoos
patch -p0 <$patch_folder/speculoos.patch
cd Speculoos/src
(make || true)
cd _build
/usr/bin/ocamlc.opt ../../../ocaml-cudd/_build/cudd_stub.cmo ../../../ocaml-aiger/_build/aiger.cma ../../../ocaml-aiger/_build/aigerImperative.cma -custom ../../../ocaml-cudd/_build/cudd.o ../../../ocaml-cudd/_build/cudd.cmo ../../../ocaml-cudd/cudd-2.5.0/cudd/libcudd.a ../../../ocaml-cudd/cudd-2.5.0/util/libutil.a ../../../ocaml-cudd/cudd-2.5.0/epd/libepd.a ../../../ocaml-cudd/cudd-2.5.0/mtr/libmtr.a ../../../ocaml-cudd/cudd-2.5.0/st/libst.a aigerBdd.cmo bddVariable.cmo common.cmo boolean.cmo integer.cmo type.cmo expression.cmo exprToBdd.cmo synthesis.cmo aigerImpBdd.cmo synthesisImp.cmo speculoos.cmo parser.cmo speculoosCompiler.cmo -warn-error -a-31 -o speculoosCompiler.byte
cp speculoosCompiler.byte ~/bin/speculoos
cd ~
rm -rf Speculoos
rm -rf ocaml-cudd
rm -rf ocaml-aiger

# install NuSMV
firefox http://nusmv.fbk.eu/NuSMV/download/getting_src-v2.html &
echo "Please download NuSMV from the displayed page"
echo "Afterwards, there should be a file 'NuSMV-2.6.0.tar.gz' in the 'Downloads' folder"
read -p "Then, press [Enter] to continue with the installation..."
cd ~
tar -xzf Downloads/NuSMV-2.6.0.tar.gz
patch -p0 <$patch_folder/nusmv_minisat.patch
patch -p0 <$patch_folder/nusmv_cudd.patch
cd NuSMV-2.6.0/NuSMV
mkdir build
cd build
cmake ..
make
cp bin/ltl2smv ~/bin
cd ~
rm -r NuSMV-2.6.0

# install nuXmv
firefox https://es-static.fbk.eu/tools/nuxmv/index.php?n=Download.Download &
echo "Please download nuXmv from the displayed page"
echo "Afterwards, there should be a file 'nuXmv-1.1.1-linux64.tar.gz' in the 'Downloads' folder"
read -p "Then, press [Enter] to continue with the installation..."
cd ~
tar -xzf Downloads/nuXmv-1.1.1-linux64.tar.gz nuXmv-1.1.1-Linux/bin/nuXmv
cp nuXmv-1.1.1-Linux/bin/nuXmv ~/bin
rm -r nuXmv-1.1.1-Linux

# install aiger tools
wget http://fmv.jku.at/aiger/aiger-1.9.4.tar.gz
tar -xzf aiger-1.9.4.tar.gz
cd aiger-1.9.4
./configure
make
cp aigtoaig smvtoaig ../bin
cd ~
rm -r aiger-1.9.4 aiger-1.9.4.tar.gz

# install MeMin
wget http://embedded.cs.uni-saarland.de/tools/MeMin/MeMin -O ~/bin/MeMin
chmod +x ~/bin/MeMin

# fix java certificates (for Ubuntu 17.10) and install newer java for Ubuntu 16.04
cd ~
wget http://download.java.net/java/GA/jdk9/9.0.4/binaries/openjdk-9.0.4_linux-x64_bin.tar.gz
tar -xzf openjdk-9.0.4_linux-x64_bin.tar.gz
if grep -q 'Ubuntu 16\.04' /etc/lsb-release 2>/dev/null; then
    # install newer Java
    mkdir -p /usr/lib/jvm
    sudo cp -r jdk-9.0.4 /usr/lib/jvm
    sudo update-alternatives --install /usr/bin/java java /opt/jdk/jdk-9.0.4/bin/java 1
    sudo update-alternatives --install /usr/bin/javac javac /opt/jdk/jdk-9.0.4/bin/javac 1
fi
if grep -q 'Ubuntu 17\.10' /etc/lsb-release 2>/dev/null; then
    # fix certificates
    sudo cp jdk-9.0.4/lib/security/cacerts /usr/lib/jvm/java-9-openjdk-amd64/lib/security
fi
rm -rf jdk-9.0.4 openjdk-9.0.4_linux-x64_bin.tar.gz

# install owl
cd ~
git clone git@gitlab.lrz.de:i7/owl.git
cd owl
./gradlew jar
cp build/libs/owl-1.2.0-SNAPSHOT.jar lib
for jar in jhoafparser-1.1.1.jar jbdd-0.2.0.jar guava-23.4-jre.jar naturals-util-0.8.0.jar fastutil-8.1.1.jar commons-cli-1.4.jar antlr4-runtime-4.7.jar; do
    find ~/.gradle -name "$jar" -exec cp {} lib \;
done
cd ..

# install strix
cd ~
git clone git@gitlab.lrz.de:i7/strix.git
cd strix
make
cp bin/strix ~/bin
cp bin/kisstospec ~/bin
cp run_strix.sh ~/bin
cd ~

# install spot
cd ~
wget http://www.lrde.epita.fr/dload/spot/spot-2.5.tar.gz
tar -xzf spot-2.5.tar.gz
cd spot-2.5
./configure --disable-python
make
cp bin/ltlsynt ~/bin
cd ~

set +x

echo "Everything went fine!"
echo "Use the 'strix' command to invoke the base synthesis engine"
echo "Use the 'run_strix.sh' command to invoke the complete toolchain"
echo "Both commands can be used with '--help' to display usage options"
