#!/bin/bash

#default arguments

timelimit=3600
memorylimit=8
benchmark=false
out_dir=/tmp
output_file=/dev/null
result_file=/dev/null
output_aiger=false
output_mealy=false
output_min_mealy=false
output_spec=false
output_min_spec=false
mealy_file=''
min_mealy_file=''
synthesis=true
verification=false
verbose=false
show_help=false

#parse command line arguments
POSITIONAL=()
while [[ $# -gt 0 ]]
do
key="$1"

case $key in
    -t|--timelimit)
    timelimit="$2"
    shift # past argument
    shift # past value
    ;;
    -m|--memorylimit)
    memorylimit="$2"
    shift # past argument
    shift # past value
    ;;
    -b|--benchmark)
    benchmarking=true
    result_file=$2
    shift # past argument
    shift # past value
    ;;
    -o|--output)
    out_dir=$2
    shift # past argument
    shift # past value
    ;;
    -a|--aiger)
    output_aiger=true
    shift # past argument
    ;;
    -k|--mealy)
    output_mealy=true
    shift # past value
    ;;
    --min-mealy)
    output_min_mealy=true
    shift # past argument
    ;;
    -s|--spec)
    output_spec=true
    shift # past value
    ;;
    --min-spec)
    output_min_spec=true
    shift # past argument
    ;;
    -r|--realizability)
    synthesis=false
    shift # past argument
    ;;
    -y|--verify)
    verification=true
    shift # past argument
    ;;
    -v|--verbose)
    verbose=true
    shift # past argument
    ;;
    -h|--help)
    show_help=true
    shift # past argument
    ;;
    *)    # unknown option
    POSITIONAL+=("$1") # save it in an array for later
    shift # past argument
    ;;
esac
done
set -- "${POSITIONAL[@]}" # restore positional parameters

if [ "$show_help" == true ]; then
    echo "Usage: $0 [FLAGS] [OPTIONS] <TLSF_FILE>"
    echo "FLAGS:"
    echo "    -h, --help                Display this help"
    echo "    -r, --realizability       Only check realizability"
    echo "    -y, --verify              Verify the result with the nuXmv model checker"
    echo "    -v, --verbose             Verbose output"
    echo "    -a, --aiger               Output the controller as an AIGER circuit"
    echo "    -k, --mealy               Output the Mealy machine in KISS format"
    echo "        --min-mealy           Output the minimized Mealy controller in KISS format"
    echo "    -s, --spec                Output the Mealy machine in Speculoos format"
    echo "        --min-spec            Output the minimized Mealy machine in Speculoos format"
    echo "OPTIONS:"
    echo "    -o, --output <OUT_DIR>    Output all intermediate files to OUT_DIR"
    echo "    -b, --benchmark <RESULTS> Output benchmarking information to RESULTS"
    echo "    -m, --memorylimit <LIMIT> Enforce a memory limit of LIMIT GB"
    echo "    -t, --timelimit <LIMIT>   Enforce a time limit of LIMIT seconds"
    echo "ARGS:"
    echo "    <TLSF_FILE>               Sets the input file in tlsf format to use"
    exit 0
fi

if [[ ${#POSITIONAL[@]} -eq 0 ]]; then
    echo "Error: no input file specified"
    exit 1
fi
input_file=${POSITIONAL[0]}

BASE_FILE=$(basename ${input_file%.tlsf})

# output files
BASIC_TLSF_FILE=$out_dir/$BASE_FILE.basic.tlsf
RESULT_AIG_FILE=$out_dir/$BASE_FILE.controller.aig
RESULT_AAG_FILE=$out_dir/$BASE_FILE.controller.aag
KISS_FILE=$out_dir/$BASE_FILE.kiss
MIN_KISS_FILE=$out_dir/$BASE_FILE.min.kiss
SPEC_FILE=$out_dir/$BASE_FILE.spec
MIN_SPEC_FILE=$out_dir/$BASE_FILE.min.spec
AAG_FILE=$out_dir/$BASE_FILE.aag
MIN_AAG_FILE=$out_dir/$BASE_FILE.min.aag
AIG_FILE=$out_dir/$BASE_FILE.aig
MIN_AIG_FILE=$out_dir/$BASE_FILE.min.aig
RED_AIG_FILE=$out_dir/$BASE_FILE.red.aig
MIN_RED_AIG_FILE=$out_dir/$BASE_FILE.min.red.aig
SMV_FILE=$out_dir/$BASE_FILE.smv
MONITOR_FILE=$out_dir/$BASE_FILE.monitor.aag
COMBINED_FILE=$out_dir/$BASE_FILE.combined.aag

#calucate memory limits from GB to byte
mem_soft=$(($memorylimit * 1024 * 1024))
mem_hard=$(($mem_soft + 1024))

# tool paths
STRIX=strix
STRIX_OPTIONS='-p --threads 8'
if [ "$verbose" == true ]; then
    STRIX_OPTIONS="$STRIX_OPTIONS --print-tree --print-product-states --timing"
fi

SYFCO=syfco
SYFCO_OPTIONS='-os mealy'
SMVTOAIG=smvtoaig
AIGTOAIG=aigtoaig
SPECULOOS=speculoos
COMBINEAIGER=combine-aiger
ABC=abc
NUXMV=nuXmv
LTL2SMV=ltl2smv
MEMIN=MeMin
KISSTOSPEC=kisstospec

runtime=0
runresult=''

function run_strix {
    INPUT=$1
    OUTPUT=$2
    OUT_FILE=$out_dir/$(basename $2).out
    runtime="$(date +%s%N)"
    (
        set -o pipefail
        ulimit -S -v $mem_soft
        ulimit -H -v $mem_hard
        timeout $timelimit $STRIX $STRIX_OPTIONS -i tlsf $INPUT -o $OUTPUT -f kiss 2>&1 | tee $OUT_FILE
    ) 2>/dev/null
    result=$?
    runtime=$(($(date +%s%N)-runtime))
    runtime=$(bc -l <<< "scale=2; $runtime/1000000000")
    if [[ $result -eq 0 ]]; then
        if grep -q "^REALIZABLE$" $OUT_FILE; then
            runresult='realizable';
        elif grep -q "^UNREALIZABLE$" $OUT_FILE; then
            runresult='unrealizable';
        else
            runresult='unknown'
        fi
    elif [[ $result -eq 124 ]] || [[ result -eq 137 ]]; then
        runresult='timeout'
    else
        runresult='error'
    fi
    return $result
}

function run_memin {
    if [ "$verbose" == true ]; then
        echo "Minimizing Mealy automaton"
    fi
    INPUT=$1
    OUTPUT=$2
    OUT_FILE=$out_dir/$(basename $2).out
    runtime="$(date +%s%N)"
    (
        set -o pipefail
        ulimit -S -v $mem_soft
        ulimit -H -v $mem_hard
        if [ "$verbose" == true ]; then
            timeout $timelimit $MEMIN -v 1 $INPUT 2>&1 | tee $OUT_FILE
        else
            timeout $timelimit $MEMIN -v 1 $INPUT >$OUT_FILE 2>&1
        fi
        head -n 2 $INPUT >$OUTPUT
        cat result.kiss >>$OUTPUT
        rm result.kiss
    ) 2>/dev/null
    result=$?
    runtime=$(($(date +%s%N)-runtime))
    runtime=$(bc -l <<< "scale=2; $runtime/1000000000")
    if [[ result -eq 0 ]]; then
        runresult='good'
    elif [[ result -eq 124 ]] || [[ result -eq 137 ]]; then
        runresult='timeout'
    else
        runresult='error'
    fi
    return $result
}

function run_speculoos {
    if [ "$verbose" == true ]; then
        echo "Converting Mealy automaton to AIGER circuit"
    fi
    INPUT=$1
    OUTPUT=$2
    OUT_FILE=$out_dir/$(basename $2).out
    runtime="$(date +%s%N)"
    (
        set -o pipefail
        ulimit -S -v $mem_soft
        ulimit -H -v $mem_hard
        # increase stack size to avoid stack overflows
        ulimit -s unlimited
        if [ "$verbose" == true ]; then
            timeout $timelimit $SPECULOOS $INPUT -o $OUTPUT 2>&1 | tee $OUT_FILE
        else
            timeout $timelimit $SPECULOOS $INPUT -o $OUTPUT >$OUT_FILE 2>&1
        fi
    ) 2>/dev/null
    result=$?
    runtime=$(($(date +%s%N)-runtime))
    runtime=$(bc -l <<< "scale=2; $runtime/1000000000")
    if [[ result -eq 0 ]]; then
        runresult='good'
    elif [[ result -eq 124 ]] || [[ result -eq 137 ]]; then
        runresult='timeout'
    else
        runresult='error'
    fi
    return $result
}

function size_of_aig {
    size=$(head -n 1 $1 | cut -d ' ' -f 6)
}

function size_of_kiss {
    size=$(head -n 7 $1 | grep -e '^\.s' | cut -d ' ' -f 2)
}

function reduce_aig {
    if [ "$verbose" == true ]; then
        echo "Compressing AIGER circuit"
    fi
    INPUT=$1
    OUTPUT=$2
    TMP_OUTPUT=$out_dir/$(basename $2).tmp.aig
    OUT_FILE=$out_dir/$(basename $2).out
    runtime="$(date +%s%N)"
    >$OUT_FILE
    cp $INPUT $OUTPUT
    size_of_aig $OUTPUT
    cur_size=$size
    reduction=true
    while $reduction; do
        if [[ $cur_size -gt 0 ]]; then
            # possible commands to add: drf > drw > dc2 > drwsat > dfraig
            # test refactor -z; rewrite -z interleaved with blance
            # test compress2rs = balance -l; resub -K 6 -l; rewrite -l; resub -K 6 -N 2; refactor -l; resub -K 8 -l; balance -l; resub -K 8 -N 2 -l; rewrite -l; resub -K 10 -l; rewrite -z -l; resub -K 10 -N 2 -l; balance -l; resub -K 12 -l; refactor -z -l; resub -K 12 -N 2 -l; balance -l; rewrite -z -l; balance -l '
            # test resyn2rs = balance; resub -K 6; rewrite; resub -K 6 -N 2; refactor; resub -K 8; balance; resub -K 8 -N 2; rewrite; resub -K 10; rewrite -z; resub -K 10 -N 2; balance; resub -K 12; refactor -z; resub -K 12 -N 2; rewrite -z; balance
            compress='balance -l; resub -K 6 -l; rewrite -l; resub -K 6 -N 2; refactor -l; resub -K 8 -l; balance -l; resub -K 8 -N 2 -l; rewrite -l; resub -K 10 -l; rewrite -z -l; resub -K 10 -N 2 -l; balance -l; resub -K 12 -l; refactor -z -l; resub -K 12 -N 2 -l; balance -l; rewrite -z -l; balance -l'
            if [ "$verbose" == true ]; then
                echo "read_aiger $OUTPUT; $compress; write_aiger -s $TMP_OUTPUT; quit" | $ABC 2>&1 | tee -a $OUT_FILE
            else
                echo "read_aiger $OUTPUT; $compress; write_aiger -s $TMP_OUTPUT; quit" | $ABC >>$OUT_FILE 2>&1
            fi
            result=$?
            if [[ $result -ne 0 ]]; then
                reduction=false
                return 1
            else
                size_of_aig $TMP_OUTPUT
                new_size=$size
                if [[ $new_size -lt $cur_size ]]; then
                    reduction=true
                    cp $TMP_OUTPUT $OUTPUT
                    cur_size=$new_size
                else
                    reduction=false
                fi
            fi
        else
            reduction=false
        fi
    done
    rm -f $TMP_OUTPUT
    runtime=$(($(date +%s%N)-runtime))
    runtime=$(bc -l <<< "scale=2; $runtime/1000000000")
}

function run_nuxmv_single {
    INPUT=$1
    OUT_FILE=$out_dir/$(basename $2).out
    COMMAND=$3
    (
        set -o pipefail
        ulimit -S -v $mem_soft
        ulimit -H -v $mem_hard
        if [ "$verbose" == true ]; then
            echo "read_aiger_model -i $INPUT; $COMMAND; quit" | timeout $timelimit $NUXMV -int 2>&1 | tee $OUT_FILE
        else
            echo "read_aiger_model -i $INPUT; $COMMAND; quit" | timeout $timelimit $NUXMV -int >$OUT_FILE 2>&1
        fi
    ) 2>/dev/null
    if [[ $result -eq 0 ]]; then
        if grep -q "specification.*is true" $OUT_FILE; then
            return 0
        elif grep -q "specification.*is false" $OUT_FILE; then
            return 1
        else
            return 2
        fi
    fi
}

function run_nuxmv_parallel {
    if [ "$verbose" == true ]; then
        echo "Verifying controller"
    fi
    INPUT=$1
    OUT_FILE=$1.out
    runtime="$(date +%s%N)"

    #run_nuxmv_single $INPUT $OUT_FILE.1 'encode_variables; build_boolean_model; bmc_setup; check_ltlspec_bmc_inc' & PID0=$!
    run_nuxmv_single $INPUT $OUT_FILE.2 'encode_variables; build_boolean_model; check_ltlspec_klive' & PID1=$!
    run_nuxmv_single $INPUT $OUT_FILE.3 'go; check_ltlspec' & PID2=$!

    wait $PID1
    r1=$?
    if [[ $r1 -eq 0 ]]; then
        runresult='passed';
        kill $PID2 2>/dev/null
    elif [[ $r1 -eq 1 ]]; then
        runresult='failed';
        kill $PID2 2>/dev/null
    else
        wait $PID2
        r2=$?
        if [[ $r2 -eq 0 ]]; then
            runresult='passed';
        elif [[ $r2 -eq 1 ]]; then
            runresult='failed';
        else
            runresult='unknown';
        fi
    fi
    # TODO: find better option to stop parallel process
    killall $NUXMV 2>/dev/null
    runtime=$(($(date +%s%N)-runtime))
    runtime=$(bc -l <<< "scale=2; $runtime/1000000000")
}

function construct_circuit {
    if ! $KISSTOSPEC $KISS_FILE $SPEC_FILE; then
        return 1
    fi
    if ! run_speculoos $SPEC_FILE $AAG_FILE; then
        return 1
    fi
    if ! $AIGTOAIG $AAG_FILE $AIG_FILE; then
        return 1
    fi
    if ! reduce_aig $AIG_FILE $RED_AIG_FILE; then
        return 1
    fi
    return 0
}

function construct_min_circuit {
    if ! run_memin $KISS_FILE $MIN_KISS_FILE; then
        return 1
    fi
    if ! $KISSTOSPEC $MIN_KISS_FILE $MIN_SPEC_FILE; then
        return 2
    fi
    if ! run_speculoos $MIN_SPEC_FILE $MIN_AAG_FILE; then
        return 3
    fi
    if ! $AIGTOAIG $MIN_AAG_FILE $MIN_AIG_FILE; then
        return 4
    fi
    if ! reduce_aig $MIN_AIG_FILE $MIN_RED_AIG_FILE; then
        return 5
    fi
    return 0
}

function construct_circuit_parallel {
    runtime_circuit_parallel="$(date +%s%N)"

    construct_circuit & PID1=$!
    construct_min_circuit & PID2=$!
    wait $PID1
    r1=$?
    wait $PID2
    r2=$?

    runtime_circuit_parallel=$(($(date +%s%N)-runtime_circuit_parallel))
    runtime_circuit_parallel=$(bc -l <<< "scale=2; $runtime_circuit_parallel/1000000000")

    if [ -f $MIN_KISS_FILE ]; then
        size_of_kiss $MIN_KISS_FILE
        size_min_mealy=$size
    fi
    if [ -f $AIG_FILE ]; then
        size_of_aig $AIG_FILE
        size_aig=$size
    fi
    if [ -f $MIN_AIG_FILE ]; then
        size_of_aig $MIN_AIG_FILE
        size_min_aig=$size
    fi
    if [ -f $RED_AIG_FILE ]; then
        size_of_aig $RED_AIG_FILE
        size_red_aig=$size
    fi
    if [ -f $MIN_RED_AIG_FILE ]; then
        size_of_aig $MIN_RED_AIG_FILE
        size_min_red_aig=$size
    fi

    if [[ $r1 -ne 0 ]]; then
        result_circuit_parallel='error'
        return 1
    fi
    if [[ $r2 -ne 0 ]]; then
        result_circuit_parallel='error'
        return 1
    fi

    if [[ $size_red_aig -le $size_min_red_aig ]]; then
        $AIGTOAIG $RED_AIG_FILE $RESULT_AIG_FILE
        result_circuit_parallel=$size_red_aig
    else
        $AIGTOAIG $MIN_RED_AIG_FILE $RESULT_AIG_FILE
        result_circuit_parallel=$size_min_red_aig
    fi
}

function run_realizability_toolchain {
	if ! $SYFCO $SYFCO_OPTIONS -f basic $input_file >$BASIC_TLSF_FILE; then
        result_realizability='error_syfco'
        return 1
    fi

    run_strix $BASIC_TLSF_FILE $KISS_FILE
    time_realizability=$runtime
    result_realizability=$runresult
}

function run_synthesis_toolchain {
    if [[ $result_realizability == 'realizable' ]]; then
        size_of_kiss $KISS_FILE
        size_mealy=$size

        construct_circuit_parallel

        result_synthesis=$result_circuit_parallel
        time_synthesis=$(bc -l <<< "scale=2; $time_realizability + $runtime_circuit_parallel")
        time_seconds=$(LC_NUMERIC="en_US.UTF-8" printf "%.0f" "$time_synthesis")
        if [[ $time_seconds -ge $timelimit ]]; then
            result_synthesis='timeout'
        fi
    elif [[ $result_realizability == 'unrealizable' ]]; then
        result_synthesis=0
        time_synthesis=$time_realizability
    else
        result_synthesis=$result_realizability
        time_synthesis=$time_realizability
    fi
}

function verify_result {
    if [[ $result_synthesis != 'error' ]] && [[ $result_synthesis != 'timeout' ]] && [ -f $RESULT_AIG_FILE ]; then
        if ! $SYFCO $SYFCO_OPTIONS -f smv -m fully $input_file >$SMV_FILE; then
            result_verification='error_tlsf_to_smv'
            return 1
        fi
        if ! $SMVTOAIG -L $LTL2SMV $SMV_FILE >$MONITOR_FILE 2>/dev/null; then
            result_verification='error_smv_to_aig'
            return 1
        fi
        if ! $COMBINEAIGER $MONITOR_FILE $RESULT_AIG_FILE >$COMBINED_FILE; then
            result_verification='error_combine_aiger'
            return 1
        fi
        run_nuxmv_parallel $COMBINED_FILE
        time_verification=$runtime
        result_verification=$runresult
        if [[ "$result_verification" == 'passed' ]]; then
            echo "Controller is verified!"
        elif [[ "$result_verification" == 'failed' ]]; then
            echo "Counterexample for controller found!"
        else
            echo "Controller could not be verified within the time limit!"
        fi
    fi
}

if [ "$benchmarking" == true ]; then
    echo -n "$(basename $input_file)" >>$result_file
fi

# test realizability
run_realizability_toolchain

if [ "$benchmarking" == true ]; then
    echo -n ",$result_realizability,$time_realizability" >>$result_file
fi

if [ "$synthesis" == true ]; then
    run_synthesis_toolchain
    if [ "$benchmarking" == true ]; then
        echo -n ",$result_synthesis,$time_synthesis,$size_mealy,$size_min_mealy,$size_aig,$size_min_aig,$size_red_aig,$size_min_red_aig" >>$result_file
    else
        if [ "$output_aiger" == true ]; then
            if [[ $result_synthesis != 'error' ]] && [[ $result_synthesis != 'timeout' ]] && [ -f $RESULT_AIG_FILE ]; then
                $AIGTOAIG $RESULT_AIG_FILE $RESULT_AAG_FILE
                sed -n '/^c/q;p' $RESULT_AAG_FILE
            else
                echo "Could not produce an AIGER circuit"
            fi
        fi
        if [ "$output_mealy" == true ]; then
            if [ -f $KISS_FILE ]; then
                cat $KISS_FILE
            else
                echo "Could not produce an Mealy machine"
            fi
        fi
        if [ "$output_min_mealy" == true ]; then
            if [ -f $MIN_KISS_FILE ]; then
                cat $MIN_KISS_FILE
            else
                echo "Could not produce a minimized Mealy machine"
            fi
        fi
        if [ "$output_spec" == true ]; then
            if [ -f $SPEC_FILE ]; then
                cat $SPEC_FILE
            else
                echo "Could not produce an Mealy machine"
            fi
        fi
        if [ "$output_min_spec" == true ]; then
            if [ -f $MIN_SPEC_FILE ]; then
                cat $MIN_SPEC_FILE
            else
                echo "Could not produce a minimized Mealy machine"
            fi
        fi
    fi
    if [ "$verification" == true ]; then
        verify_result
        if [ "$benchmarking" == true ]; then
            echo -n ",$result_verification,$time_verification" >>$result_file
        fi
    fi
fi

if [ "$benchmarking" == true ]; then
    echo >>$result_file
fi
