#include "MealyMachine.h"

#include <iomanip>
#include <bitset>
#include <cmath>

namespace controller {

state_id_t MealyMachine::addState() {
    const state_id_t state = n_states;
    n_states++;
    n_transitions += n_inputs;
    successors.resize(successors.size() + n_inputs);
    return state;
}

state_id_t MealyMachine::getNumberOfStates() const {
    return n_states;
}

void MealyMachine::setSuccessor(const state_id_t s, const ltl::letter_t input, const state_id_t successor, const std::vector<ltl::letter_t>& outputs) {
    ternary_int<ltl::letter_t> output;
    if (outputs.size() > 1) {
        output = quine_minimal_implicant(tlsf.n_outputs, outputs);
    }
    else {
        output = ternary_int<ltl::letter_t>(outputs[0]);
    }

    successors[s * n_inputs + input] = Successor(successor, output);
}

void MealyMachine::reduce() {
    // identify states that don't do anything
    std::vector<bool> top_state(n_states, true);
    const ternary_int<ltl::letter_t> any_output = true_clause<ltl::letter_t>(tlsf.n_outputs);

    for (state_id_t q = 0; q < n_states; ++q) {
        state_id_t k = q * n_inputs;
        for (ltl::letter_t i = 0; i < n_inputs; ++i, ++k) {
            if (successors[k].successor != q || successors[k].output != any_output) {
                top_state[q] = false;
                break;
            }
        }
    }

    // identify transitions that don't do anything
    state_id_t k = 0;
    for (state_id_t q = 0; q < n_states; ++q) {
        for (ltl::letter_t i = 0; i < n_inputs; ++i, ++k) {
            if (top_state[successors[k].successor] && successors[k].output == any_output) {
                successors[k].successor = TOP_STATE;
                n_transitions--;
            }
        }
    }
}

void MealyMachine::print_spec(std::ostream& out) const {
    if (n_states > 1) {
        size_t log_n_states = (state_id_t)ceil(log2(n_states));
        out << "var state : int " << log_n_states << ";" << std::endl;
    }
    for (const auto& input : tlsf.inputs) {
        out << "var " << input << " : bool;" << std::endl;
    }
    for (const auto& output : tlsf.outputs) {
        out << "var " << output << " : bool;" << std::endl;
    }

    state_id_t k = 0;
    for (state_id_t q = 0; q < n_states; ++q) {
        out << std::endl;
        for (ltl::letter_t i = 0; i < n_inputs; ++i, ++k) {
            state_id_t successor = successors[k].successor;

            if (successor != TOP_STATE) {
                if (n_states > 1 || tlsf.n_inputs > 0) {
                    out << "when ( ";
                    for (ltl::letter_t input = 0; input < tlsf.n_inputs; ++input) {
                        if ((i & (1 << input)) == 0) {
                            out << "!";
                        }
                        else {
                            out << " ";
                        }
                        out << tlsf.inputs[input];
                        if (n_states > 1 || input + 1 < tlsf.n_inputs) {
                            out << " & ";
                        }
                    }
                    if (n_states > 1) {
                        out << "state = " << q;
                    }
                    out << " ) { ";
                }
                if (n_states > 1) {
                    out << "state <- " << successor << ";";
                }
                ternary_int<ltl::letter_t> o = successors[k].output;
                for (ltl::letter_t output = 0; output < tlsf.n_outputs; ++output) {
                    const int bit = 1 << output;
                    if ((o.dashes & bit) == 0) {
                        out << tlsf.outputs[output] << " <- ";
                        if ((o.number & bit) != 0) {
                            out << "1";
                        }
                        else {
                            out << "0";
                        }
                        out << "; ";
                    }
                }
                if (n_states > 1 || tlsf.n_inputs > 0) {
                    out << "}";
                }
                out << std::endl;
            }
        }
    }
}

void MealyMachine::print_kiss(std::ostream& out) const {
    out << ".inputs";
    for (const auto& input : tlsf.inputs) {
        out << " " << input;
    }
    out << std::endl;
    out << ".outputs";
    for (const auto& output : tlsf.outputs) {
        out << " " << output;
    }
    out << std::endl;
    out << ".i " << tlsf.n_inputs << std::endl;
    out << ".o " << tlsf.n_outputs << std::endl;
    out << ".p " << n_transitions << std::endl;
    out << ".s " << n_states << std::endl;
    out << ".r S0" << std::endl;
    state_id_t k = 0;
    for (state_id_t q = 0; q < n_states; ++q) {
        for (ltl::letter_t i = 0; i < n_inputs; ++i, ++k) {
            state_id_t successor = successors[k].successor;
            if (successor != TOP_STATE) {
                for (ltl::letter_t input = 0; input < tlsf.n_inputs; ++input) {
                    if ((i & (1 << input)) == 0) {
                        out << "0";
                    }
                    else {
                        out << "1";
                    }
                }
                out << " S" << q << " S" << successor << " ";
                ternary_int<ltl::letter_t> o = successors[k].output;
                for (ltl::letter_t output = 0; output < tlsf.n_outputs; ++output) {
                    const int bit = 1 << output;
                    if ((o.dashes & bit) == 0) {
                        if ((o.number & bit) != 0) {
                            out << "1";
                        }
                        else {
                            out << "0";
                        }
                    }
                    else {
                        out << "-";
                    }
                }
                out << std::endl;
            }
        }
    }
}

}
