#include <iostream>
#include <vector>
#include <string>
#include <set>
#include <cassert>

namespace controller {

// count the number of set bits in a number
template <typename I>
inline int popcount(I number) {
    int count = 0;
    while (number > 0) {
        if (number % 2 != 0) {
            count++;
        }
        number >>= 1;
    }
    return count;
    // TODO: use builtin popcount function if available
    // return __builtin_popcount(number);
}

// tests if number x is a power of two
template <typename I>
inline bool is_power_of_two(const I& x) {
    return x && !(x & (x-1));
}

template <typename I>
struct ternary_int {
    I number;
    I dashes;

    ternary_int() : number(0), dashes(0) {}
    ternary_int(I n) : number(n), dashes(0) {}
    ternary_int(I n, I d) : number(n), dashes(d) {}

    int num_ones() {
        return popcount(number);
    }

    int num_dashes() {
        return popcount(dashes);
    }

    // comparison operators
    friend bool operator==(const ternary_int<I>& a, const ternary_int<I>& b) {
        return a.number == b.number && a.dashes == b.dashes;
    }
    friend bool operator!=(const ternary_int<I>& a, const ternary_int<I>& b) {
        return !(a == b);
    }
    friend bool operator<(const ternary_int<I>& a, const ternary_int<I>& b) {
        return a.number < b.number || (a.number == b.number && a.dashes < b.dashes);
    }
    friend bool operator<=(const ternary_int<I>& a, const ternary_int<I>& b) {
        return a.number < b.number || (a.number == b.number && a.dashes <= b.dashes);
    }
    friend bool operator>(const ternary_int<I>& a, const ternary_int<I>& b) {
        return a.number > b.number || (a.number == b.number && a.dashes > b.dashes);
    }
    friend bool operator>=(const ternary_int<I>& a, const ternary_int<I>& b) {
        return a.number > b.number || (a.number == b.number && a.dashes >= b.dashes);
    }

    // print operator
    friend std::ostream &operator<<(std::ostream &out, const ternary_int<I>& b) {
        std::vector<int> bits;
        constexpr int num_bits = 8;
        int count = 0;
        I n = b.number;
        I d = b.dashes;

        while (n > 0 || count < num_bits) {
            if (d%2 == 0) {
                bits.push_back(n%2);
            }
            else {
                bits.push_back(2);
            }

            n >>= 1;
            d >>= 1;
            count++;
        }

        for (auto rit = bits.rbegin(); rit != bits.rend(); ++rit) {
            if (*rit == 2) {
                out << "-";
            }
            else {
                out << *rit;
            }
        }
        return out;
    }

    // combine lhs and rhs if they differ only in one bit
    // return true if combination is possible, otherwise return false
    bool combine(const ternary_int<I>& lhs, const ternary_int<I>& rhs) {
        if (lhs.dashes == rhs.dashes) {
            const I tmp_dashes = lhs.number ^ rhs.number;
            if (is_power_of_two(tmp_dashes)) {
                number = lhs.number & rhs.number;
                dashes = tmp_dashes | lhs.dashes;
                return true;
            }
        }
        return false;
    }
};

template <typename I>
ternary_int<I> true_clause(const int num_bits) {
    return ternary_int<I>(0, (1 << num_bits) - 1);
}

//print the table
template<typename I>
void print_table(const std::vector< std::vector< std::set< ternary_int<I> > > >& table, size_t max_size) {
    for (size_t size = 0; size <= max_size; ++size) {
        std::cout<<"\n-------------------------------------"<<std::endl;
        std::cout << "Size " << size << ":" << std::endl;
        for (size_t i = 0; i < table[size].size(); ++i) {
            if (table[size][i].size() > 0) {
                std::cout << i;
            }
            for (const ternary_int<I>& number : table[size][i]) {
                std::cout << "\t\t" << number << std::endl;
            }
        }
        std::cout<<"\n-------------------------------------"<<std::endl;
    }
}

template<typename I>
ternary_int<I> quine_minimal_implicant(const int num_bits, const std::vector<I>& values) {
    assert(num_bits > 0);
    assert(values.size() > 0);

    if (values.size() == ((size_t)1 << num_bits)) {
        // Assuming all values are distinct, all values can be combined.
        // Directly return minimal implicant for this number of bits.
        return true_clause<I>(num_bits);
    }

    // table of minterms and implicants
    // first dimension: number of dashes
    // second dimension: number of ones
    std::vector< std::vector< std::set< ternary_int<I> > > > table;

    // resize table according to number of bits
    table.resize(num_bits + 1);
    for (size_t size = 0; size < table.size(); ++size) {
        table[size].resize(num_bits + 1);
    }

    // fill table with input values
    for (const auto in : values) {
        ternary_int<I> b_in(in);
        const int num_ones = b_in.num_ones();
        table[0][num_ones].insert(b_in);
    }

    // compute implicants by increasing size
    bool new_value = true;
    size_t size = 0;
    while (new_value) {
        new_value = false;
        for (size_t i = 0; i+1 < table[size].size(); ++i) {
            for (const ternary_int<I>& lhs : table[size][i]) {
                for (const ternary_int<I>& rhs : table[size][i+1]) {
                    ternary_int<I> result;
                    if (result.combine(lhs, rhs)) {
                        table[size + 1][i].insert(result);
                        new_value = true;
                    }
                }
            }
        }
        if (new_value) {
            size++;
        }
    }

    for (size_t i = 0; i+1 < table[size].size(); ++i) {
        if (!table[size][i].empty()) {
            return *table[size][i].begin();
        }
    }

    throw std::invalid_argument("empty set of implicants");
}

}
