#include "Parser.h"

#include <iostream>
#include <fstream>
#include <sstream>

namespace ltl {

Parser::Parser(const owl::OwlThread& owl) : owl(owl) { }

Parser::~Parser() { }

TLSF Parser::parse(const std::string& input_file, const bool print_formula) const {
    // read input file
    std::string text;

    std::ifstream infile(input_file);
    if (infile.is_open()) {
        text = std::string((std::istreambuf_iterator<char>(infile)), std::istreambuf_iterator<char>());
    }
    else {
        throw std::runtime_error("could not open input file");
    }
    if (infile.bad()) {
        throw std::runtime_error("could not read input file");
    }
    infile.close();

    if (print_formula) {
        std::cout << "Input file: " << std::endl << text;
    }

    // parse formula and input and output variables
    owl::FormulaFactory factory = owl.createFormulaFactory();
    owl::FormulaRewriter rewriter = owl.createFormulaRewriter();

    std::vector<std::string> inputs;
    std::vector<std::string> outputs;

    owl::Formula parsedFormula = parse_text(factory, text, inputs, outputs);
    if (print_formula) {
        std::cout << "Parsed formula:" << std::endl;
        parsedFormula.print();
        std::cout << "Inputs:";
        for (const auto& i : inputs) {
            std::cout << " " << i;
        }
        std::cout << std::endl;
        std::cout << "Outputs:";
        for (const auto& o : outputs) {
            std::cout << " " << o;
        }
        std::cout << std::endl;
    }

    // simplify formula
    owl::Formula simplifiedFormula = rewriter.simplify(parsedFormula);
    if (print_formula) {
        std::cout << "Simplified formula:" << std::endl;
        simplifiedFormula.print();
    }

    // remove unused variables
    std::map<int, int> mapping = std::map<int, int>();
    owl::Formula shiftedFormula = rewriter.shift_literals(simplifiedFormula, mapping);
    if (print_formula) {
        std::cout << "Shifted formula:" << std::endl;
        shiftedFormula.print();
    }

    size_t new_inputs_size = 0;
    size_t new_outputs_size = 0;
    if (print_formula) {
        std::cout << "Mapping:" << std::endl;
    }
    for (const auto& entry : mapping) {
        if (print_formula) {
            std::cout << entry.first << " -> " << entry.second << std::endl;
        }
        const size_t i = (size_t)entry.first;
        if (i < inputs.size()) {
            new_inputs_size++;
        }
        else {
            new_outputs_size++;
        }
    }
    std::vector<std::string> new_inputs(new_inputs_size);
    std::vector<std::string> new_outputs(new_outputs_size);
    for (const auto& entry : mapping) {
        size_t i = (size_t)entry.first;
        size_t j = (size_t)entry.second;
        if (i < inputs.size()) {
            new_inputs[j] = inputs[i];
        }
        else {
            i -= inputs.size();
            j -= new_inputs.size();
            new_outputs[j] = outputs[i];
        }
    }

    return TLSF(shiftedFormula, new_inputs, new_outputs);
}

}

