#include "OCLPGSolver.h"

namespace ocl {

OCLPGSolver::OCLPGSolver(OCLDriver& driver, pg::PGArena& arena, const bool verbose) :
    PGSolver(arena, false, verbose),
    driver(driver)
{
    driver.addSourceFile("pg/PGArenaDefinitions.h");
    driver.addSourceFile("ocl/OCLPGSolverDefinitions.h");
    driver.addSourceFile("ocl/OCLPGSolver.cl");

    driver.addConstant("N_ENV_NODES", n_env_nodes);
    driver.addConstant("N_ENV_ACTIONS", arena.n_env_actions);
    driver.addConstant("N_SYS_NODES", n_sys_nodes);

    driver.addConstant("INITIAL_NODE", arena.initial_node);

    driver.addConstant("N_COLORS", arena.n_colors);

    switch(arena.parity_type) {
        case Parity::ODD:
            driver.addDefinition("PARITY_ODD");
            break;
        case Parity::EVEN:
            driver.addDefinition("PARITY_EVEN");
            break;
        default:
            throw std::invalid_argument("Unsupported parity type");
    }

    driver.setWorkSize(std::max(n_sys_nodes, n_env_nodes));

    kernel_strategy_iteration_init = driver.addKernel("strategy_iteration_init");
    kernel_update_nodes = driver.addKernel("update_nodes");
    kernel_bellman_ford_init = driver.addKernel("bellman_ford_init");
    kernel_bellman_ford_sys_iteration = driver.addKernel("bellman_ford_sys_iteration");
    kernel_bellman_ford_env_iteration = driver.addKernel("bellman_ford_env_iteration");
    kernel_strategy_improvement = driver.addKernel("strategy_improvement");
}

OCLPGSolver::~OCLPGSolver() {
    delete kernel_strategy_iteration_init;
    delete kernel_update_nodes;
    delete kernel_bellman_ford_init;
    delete kernel_bellman_ford_sys_iteration;
    delete kernel_bellman_ford_env_iteration;
    delete kernel_strategy_improvement;
}

void OCLPGSolver::init() {
    sys_succs_begin_buffer = new compute::vector<edge_id_t>(n_sys_nodes + 1, driver.context);
    sys_succs_buffer       = new compute::vector<node_id_t>(n_sys_edges, driver.context);

    compute::copy(arena.sys_succs_begin.cbegin(), arena.sys_succs_begin.cbegin() + n_sys_nodes + 1, sys_succs_begin_buffer->begin(), driver.queue);
    compute::copy(arena.sys_succs.cbegin(), arena.sys_succs.cbegin() + n_sys_edges, sys_succs_buffer->begin(), driver.queue);

    colors_buffer = new compute::vector<color_t>(n_sys_edges, driver.context);

    compute::copy(arena.colors.cbegin(), arena.colors.cbegin() + n_sys_edges, colors_buffer->begin(), driver.queue);

    sys_distance_buffer = new compute::vector<distance_t>(n_sys_nodes * arena.n_colors, driver.context);
    env_distance_buffer = new compute::vector<distance_t>(n_env_nodes * arena.n_colors, driver.context);

    realizable_buffer = new compute::buffer(driver.context, sizeof(cl_bool));

    driver.queue.enqueue_write_buffer(*realizable_buffer, 0, sizeof(cl_bool), &realizable);

    kernel_strategy_iteration_init->setArg("sys_succs_begin", *sys_succs_begin_buffer);
    kernel_strategy_iteration_init->setArg("sys_succs", *sys_succs_buffer);
    kernel_strategy_iteration_init->setArg("sys_distances", *sys_distance_buffer);
    kernel_strategy_iteration_init->setArg("env_distances", *env_distance_buffer);

    kernel_update_nodes->setArg("sys_succs_begin", *sys_succs_begin_buffer);
    kernel_update_nodes->setArg("sys_succs", *sys_succs_buffer);
    kernel_update_nodes->setArg("sys_distances", *sys_distance_buffer);
    kernel_update_nodes->setArg("env_distances", *env_distance_buffer);
    kernel_update_nodes->setArg("realizable", *realizable_buffer);

    kernel_bellman_ford_init->setArg("sys_distances", *sys_distance_buffer);
    kernel_bellman_ford_init->setArg("env_distances", *env_distance_buffer);

    kernel_bellman_ford_sys_iteration->setArg("sys_succs_begin", *sys_succs_begin_buffer);
    kernel_bellman_ford_sys_iteration->setArg("sys_succs", *sys_succs_buffer);
    kernel_bellman_ford_sys_iteration->setArg("colors", *colors_buffer);
    kernel_bellman_ford_sys_iteration->setArg("sys_distances", *sys_distance_buffer);
    kernel_bellman_ford_sys_iteration->setArg("env_distances", *env_distance_buffer);

    kernel_bellman_ford_env_iteration->setArg("change", *driver.return_value_buffer);
    kernel_bellman_ford_env_iteration->setArg("colors", *colors_buffer); // TODO
    kernel_bellman_ford_env_iteration->setArg("sys_distances", *sys_distance_buffer);
    kernel_bellman_ford_env_iteration->setArg("env_distances", *env_distance_buffer);

    kernel_strategy_improvement->setArg("change", *driver.return_value_buffer);
    kernel_strategy_improvement->setArg("sys_succs_begin", *sys_succs_begin_buffer);
    kernel_strategy_improvement->setArg("sys_succs", *sys_succs_buffer);
    kernel_strategy_improvement->setArg("colors", *colors_buffer);
    kernel_strategy_improvement->setArg("sys_distances", *sys_distance_buffer);
    kernel_strategy_improvement->setArg("env_distances", *env_distance_buffer);
}

void OCLPGSolver::clean() {
    successors.resize(n_sys_edges);

    node_id_t* sys_succs = new node_id_t[n_sys_edges];
    compute::copy(sys_succs_buffer->cbegin(), sys_succs_buffer->cbegin() + n_sys_edges, sys_succs, driver.queue);

    for (edge_id_t j = 0; j < n_sys_edges; ++j) {
        successors[j] = IS_MARKED_EDGE(sys_succs[j]);
    }

    delete[] sys_succs;

    delete sys_succs_buffer;
    delete sys_succs_begin_buffer;
    delete colors_buffer;
    delete sys_distance_buffer;
    delete env_distance_buffer;

    delete realizable_buffer;
}

void OCLPGSolver::reset() { }

void OCLPGSolver::solve_game() {
    init();

    strategy_iteration();

    clean();
}

void OCLPGSolver::strategy_iteration() {
    print_debug("Initializing strategy iteration…");
	driver.executeVoidKernel(kernel_strategy_iteration_init);
    print_values_debug();

    bool change = true;
    while (change && !realizable) {
        bellman_ford();
        print_debug("Executing strategy improvement…");
        change = driver.executeBoolKernel(kernel_strategy_improvement);
        print_values_debug();
        print_debug("Marking solved nodes");
        driver.executeVoidKernel(kernel_update_nodes);
        print_values_debug();
        driver.queue.enqueue_read_buffer(*realizable_buffer, 0, sizeof(cl_bool), &realizable);
    }
}

void OCLPGSolver::bellman_ford() {
    print_debug("Executing Bellman-Ford algorithm…");
    driver.executeVoidKernel(kernel_bellman_ford_init);
    bool change = true;
    while (change) {
        print_debug("Executing Bellman-Ford sys iteration…");
        driver.executeVoidKernel(kernel_bellman_ford_sys_iteration);
        print_values_debug();
        print_debug("Executing Bellman-Ford env iteration…");
        change = driver.executeBoolKernel(kernel_bellman_ford_env_iteration);
        print_values_debug();
    }
}

inline void OCLPGSolver::print_values_debug() {
    if (verbose) {
        node_id_t* sys_succs = new node_id_t[n_sys_edges];
        distance_t* sys_distances = new distance_t[n_sys_nodes * arena.n_colors];
        distance_t* env_distances = new distance_t[n_env_nodes * arena.n_colors];

        compute::copy(sys_succs_buffer->cbegin(), sys_succs_buffer->cbegin() + n_sys_edges, sys_succs, driver.queue);
        compute::copy(sys_distance_buffer->cbegin(), sys_distance_buffer->cbegin() + (n_sys_nodes * arena.n_colors), sys_distances, driver.queue);
        compute::copy(env_distance_buffer->cbegin(), env_distance_buffer->cbegin() + (n_env_nodes * arena.n_colors), env_distances, driver.queue);

        std::cout << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;
        std::cout << "---- sys nodes -----" << std::endl;
        std::cout << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;
        std::cout << "        v     o(v)    d(v)" << std::endl;
        for (size_t i = 0; i < n_sys_nodes; ++i) {
            if (i == arena.initial_node) {
                std::cout << ">";
            }
            else {
                std::cout << " ";
            }
            std::cout << std::setw(8) << i;

            std::cout << " [";
            for (edge_id_t j = arena.sys_succs_begin[i]; j != arena.sys_succs_begin[i+1]; ++j) {
                if (IS_MARKED_EDGE(sys_succs[j])) {
                    std::cout << " " << arena.sys_succs[j];
                }
            }
            std::cout << " ]";

            for (color_t c = 0; c < arena.n_colors; c++) {
                distance_t distance = sys_distances[i*arena.n_colors + c];
                if (distance == DIST_INFINITY) {
                    std::cout << "       ∞";
                    break;
                }
                else if (distance == DIST_SOLVED) {
                    std::cout << "  solved";
                    break;
                }
                else {
                    std::cout << std::setw(8) << distance;
                }
            }
            std::cout << std::endl;
        }
        std::cout << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;
        std::cout << "---- env nodes -----" << std::endl;
        std::cout << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;
        std::cout << "        v     d(v)" << std::endl;
        for (size_t i = 0; i < n_env_nodes; ++i) {
            if (i == arena.initial_node) {
                std::cout << ">";
            }
            else {
                std::cout << " ";
            }
            std::cout << std::setw(8) << i;
            for (color_t c = 0; c < arena.n_colors; c++) {
                distance_t distance = env_distances[i*arena.n_colors + c];
                if (distance == DIST_INFINITY) {
                    std::cout << "       ∞";
                    break;
                }
                else if (distance == DIST_SOLVED) {
                    std::cout << "  solved";
                    break;
                }
                else {
                    std::cout << std::setw(8) << distance;
                }
            }
            std::cout << std::endl;
        }
        std::cout << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;

        delete[] sys_succs;
        delete[] sys_distances;
        delete[] env_distances;
    }
}

}
