#ifdef PARITY_ODD
    #define COLOR_DISTANCE_DELTA(k)  ( ((k & 1) << 1) - 1)
#else // PARITY_EVEN
    #define COLOR_DISTANCE_DELTA(k)  (-((k & 1) << 1) + 1)
#endif

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel strategy_iteration_init(
        global const edge_id_t* sys_succs_begin,
        global node_id_t* sys_succs,
        global distance_t* sys_distances,
        global distance_t* env_distances
) {
    const node_id_t i = get_global_id(0);
    const dist_id_t k = i * N_COLORS;
    if (i < N_SYS_NODES) {
        sys_distances[k] = 0;
        for (edge_id_t j = sys_succs_begin[i]; j != sys_succs_begin[i+1]; j++) {
            UNMARK_EDGE(sys_succs[j]);
        }
    }
    if (i < N_ENV_NODES) {
        env_distances[k] = 0;
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel update_nodes(
        global const edge_id_t* sys_succs_begin,
        global node_id_t* sys_succs,
        global distance_t* sys_distances,
        global distance_t* env_distances,
        global cl_bool* realizable
) {
    const node_id_t i = get_global_id(0);
    const dist_id_t k = i * N_COLORS;
    if (i < N_SYS_NODES && sys_distances[k] == DIST_INFINITY) {
        // node won by system player
        sys_distances[k] = DIST_SOLVED;
        for (edge_id_t j = sys_succs_begin[i]; j != sys_succs_begin[i+1]; j++) {
            const node_id_t succ_marked = sys_succs[j];
            if (IS_MARKED_EDGE(succ_marked)) {
                const node_id_t successor = succ_marked & MASK_UNMARKED_EDGE;
                if (env_distances[successor * N_COLORS] < DIST_INFINITY) {
                    UNMARK_EDGE(sys_succs[j]);
                }
            }
        }
    }
    if (i < N_ENV_NODES && env_distances[k] == DIST_INFINITY) {
        // node won by system player
        env_distances[k] = DIST_SOLVED;
        if (i == INITIAL_NODE) {
            *realizable = true;
        }
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel bellman_ford_init(
        global distance_t* sys_distances,
        global distance_t* env_distances
) {
    const node_id_t i = get_global_id(0);
    const dist_id_t k = i * N_COLORS;
    if (i < N_ENV_NODES && env_distances[k] != DIST_SOLVED) {
        env_distances[k] = DIST_INFINITY;
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel bellman_ford_sys_iteration(
        global const edge_id_t* sys_succs_begin,
        global const node_id_t* sys_succs,
        global const color_t* colors,
        global distance_t* sys_distances,
        global const distance_t* env_distances
) {
    const node_id_t i = get_global_id(0);
    const dist_id_t k = i * N_COLORS;
    if (i < N_SYS_NODES && sys_distances[k] != DIST_SOLVED) {
        distance_t distance[N_COLORS];

        for (dist_id_t l = 0, m = k; l < N_COLORS; l++, m++) {
            distance[l] = 0;
        }

        for (edge_id_t j = sys_succs_begin[i]; j != sys_succs_begin[i+1]; j++) {
            const node_id_t succ_marked = sys_succs[j];
            if (IS_MARKED_EDGE(succ_marked)) {
                const node_id_t successor = succ_marked & MASK_UNMARKED_EDGE;
                dist_id_t m = successor * N_COLORS;

                if (env_distances[m] < DIST_INFINITY) {
                    cl_bool local_change = false;

                    const color_t cur_color = colors[j];
                    const distance_t cur_color_change = COLOR_DISTANCE_DELTA(cur_color);
                    distance[cur_color] -= cur_color_change;

                    for (dist_id_t l = 0; l < N_COLORS; l++, m++) {
                        const distance_t d = distance[l];
                        const distance_t d_succ = env_distances[m];
                        if (local_change || d_succ > d) {
                            distance[l] = d_succ;
                            local_change = true;
                        }
                        else if (d_succ != d) {
                            break;
                        }
                    }

                    distance[cur_color] += cur_color_change;
                }
                else {
                    distance[0] = DIST_INFINITY;
                    break;
                }
            }
        }

        for (dist_id_t l = 0, m = k; l < N_COLORS; l++, m++) {
            sys_distances[m] = distance[l];
        }
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel bellman_ford_env_iteration(
        global const color_t* colors,
        global const distance_t* sys_distances,
        global distance_t* env_distances,
        global cl_bool* change
) {
    const node_id_t i = get_global_id(0);
    const dist_id_t k = i * N_COLORS;
    if (i < N_ENV_NODES && env_distances[k] != DIST_SOLVED) {
        distance_t distance[N_COLORS];

        for (dist_id_t l = 0, m = k; l < N_COLORS; l++, m++) {
            distance[l] = env_distances[m];
        }

        const node_id_t succs_begin = i * N_ENV_ACTIONS;
        const node_id_t succs_end = succs_begin + N_ENV_ACTIONS;
        for (node_id_t successor = succs_begin; successor != succs_end; ++successor) {
            dist_id_t m = successor * N_COLORS;
            if (sys_distances[m] < DIST_INFINITY) {
                cl_bool local_change = false;

                for (dist_id_t l = 0; l < N_COLORS; l++, m++) {
                    const distance_t d = distance[l];
                    const distance_t d_succ = sys_distances[m];
                    if (local_change || d_succ < d) {
                        distance[l] = d_succ;
                        local_change = true;
                    }
                    else if (d_succ != d) {
                        break;
                    }
                }
                if (local_change) {
                    *change = true;
                }
            }
        }

        for (dist_id_t l = 0, m = k; l < N_COLORS; l++, m++) {
            env_distances[m] = distance[l];
        }
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel strategy_improvement(
        global const edge_id_t* sys_succs_begin,
        global node_id_t* sys_succs,
        global const color_t* colors,
        global const distance_t* sys_distances,
        global const distance_t* env_distances,
        global cl_bool* change
) {
    const node_id_t i = get_global_id(0);
    const dist_id_t k = i * N_COLORS;
    if (i < N_SYS_NODES && sys_distances[k] < DIST_INFINITY) {
        distance_t distance[N_COLORS];
        for (dist_id_t l = 0, m = k; l < N_COLORS; l++, m++) {
            distance[l] = sys_distances[m];
        }

        for (edge_id_t j = sys_succs_begin[i]; j != sys_succs_begin[i+1]; j++) {
            UNMARK_EDGE(sys_succs[j]);
            const node_id_t successor = sys_succs[j];

            if (successor < N_ENV_NODES) {
                cl_bool improvement = true;
                const color_t cur_color = colors[j];
                const distance_t cur_color_change = COLOR_DISTANCE_DELTA(cur_color);
                distance[cur_color] -= cur_color_change;

                for (dist_id_t l = 0, m = successor * N_COLORS; l < N_COLORS; l++, m++) {
                    const distance_t d = distance[l];
                    const distance_t d_succ = env_distances[m];
                    if (d_succ > d) {
                        // strict improvement
                        *change = true;
                        break;
                    }
                    else if (d_succ != d) {
                        improvement = false;
                        break;
                    }
                }

                distance[cur_color] += cur_color_change;

                if (improvement) {
                    MARK_EDGE(sys_succs[j]);
                }
            }
        }
    }
}
