#include "PGArena.h"

#include <iostream>
#include <iomanip>
#include <sstream>
#include <cmath>
#include <limits>
#include <algorithm>
#include <cstdint>
#include <bitset>
#include <unordered_map>
#include <queue>
#include <memory>

#include <boost/functional/hash.hpp>

namespace pg {

PGArena::PGArena(const ltl::TLSF& tlsf, const owl::LabelledTree<owl::Tag, owl::Automaton>& tree) :
    tlsf(tlsf),
    ptree(ParityAutomatonTree::constructTree(tree)),
    has_lock(false),
    complete(false),
    solved(false),
    parity_type(ptree->parity_type),
    n_colors(ptree->max_color + 1),
    initial_node(0),
    n_env_actions(1 << tlsf.n_inputs),
    n_env_nodes(0),
    n_sys_edges(0)
{
    // reserve some space in the vector to avoid initial resizing, which needs locking for the parallel construction
    sys_succs_begin.reserve(4096);
    sys_succs.reserve(4096);
    sys_output.reserve(4096);
    colors.reserve(4096);
}

PGArena::~PGArena() { }

template <typename Container>
struct container_hash {
    std::size_t operator()(Container const& c) const {
        return boost::hash_range(c.cbegin(), c.cend());
    }
};

void PGArena::buildProductAutomaton(const size_t k, const bool parallel, const bool print_product_states, const bool print_product_successors) {
    product_state_t initial_state;
    ptree->getInitialState(initial_state);

    std::queue<product_state_t> queue;
    std::unordered_map< product_state_t, node_id_t, container_hash<product_state_t> > state_map;

    state_map.insert({initial_state, 0});
    queue.push(std::move(initial_state));
    sys_succs_begin.push_back(0);

    node_id_t max_env_node = 0;
    const ltl::letter_t n_sys_actions = (1 << tlsf.n_outputs);

    size_t state_length = 0;

    while (!solved && !queue.empty() && (k == 0 || n_env_nodes < k)) {

        const product_state_t state = std::move(queue.front());
        queue.pop();

        if (print_product_states) {
            std::cout << "Computing successors for " << std::setw(4) << (sys_succs_begin.size() / n_env_actions) << " = (";
            size_t si = 0;
            for (const auto s : state) {
                if (s == NODE_TOP) {
                    std::cout << "  ⊤";
                }
                else if (s == NODE_BOTTOM) {
                    std::cout << "  ⊥";
                }
                else {
                    std:: cout << " " << std::setw(2) << s;
                }
                si++;
            }
            std::cout << " )";
            if (si > state_length) {
                state_length = si;
            }
            if (state_length > si) {
                for (; si < state_length; ++si) {
                    std::cout << "   ";
                }
            }
            if (print_product_successors) {
                std::cout << " : [";
            }
            else {
                std::cout << std::endl;
            }
        }

        edge_id_t cur_n_sys_edges = 0;

        for (ltl::letter_t i = 0; i < n_env_actions; ++i) {

            // new sys node
            //std::cout << "  Input: " << std::bitset<2>(i) << std::endl;

            for (ltl::letter_t o = 0; o < n_sys_actions; ++o) {
                ltl::letter_t letter = (o << tlsf.n_inputs) + i;
                //std::cout << "   Output: " << std::bitset<2>(o) << "; Letter: " << std::bitset<4>(letter) << std::endl;

                product_state_t new_state;
                auto it = state.cbegin();
                color_t color = ptree->getSuccessor(it, new_state, letter);

                if (new_state[0] != NODE_BOTTOM) {
                    node_id_t succ;

                    auto it = state_map.find(new_state);
                    if (it == state_map.end()) {
                        succ = max_env_node + 1;
                        max_env_node++;
                        state_map.insert({new_state, succ});
                        queue.push(std::move(new_state));
                    }
                    else {
                        succ = it->second;
                    }

                    if (print_product_successors) {
                        std::cout << "  " << std::setw(2) << succ;
                        std::cout << ":" << std::setw(2) << color;
                    }

                    sys_output.push_back(o);
                    if (parallel && (sys_succs.size() == sys_succs.capacity() || colors.size() == colors.capacity())) {
                        arena_mutex.lock();
                        has_lock = true;
                    }
                    sys_succs.push_back(succ);
                    colors.push_back(color);
                    if (has_lock) {
                        arena_mutex.unlock();
                        has_lock = false;
                    }

                    cur_n_sys_edges++;
                }
                else {
                    if (print_product_successors) {
                        std::cout << "       ";
                    }
                }
            }

            if (parallel && (sys_succs_begin.size() == sys_succs_begin.capacity())) {
                arena_mutex.lock();
                has_lock = true;
            }
            sys_succs_begin.push_back(sys_succs.size());
            if (has_lock) {
                arena_mutex.unlock();
                has_lock = false;
            }
        }
        if (print_product_successors) {
            std::cout << "]" << std::endl;
        }

        if (parallel) {
            size_mutex.lock();
        }
        n_sys_edges += cur_n_sys_edges;
        n_env_nodes++;
        if (parallel) {
            size_mutex.unlock();
            change.notify_all();
        }

    }

    if (parallel) {
        size_mutex.lock();
    }
    assert(n_env_nodes * n_env_actions + 1 == sys_succs_begin.size());
    assert(n_sys_edges == sys_succs.size());
    complete = true;
    if (parallel) {
        size_mutex.unlock();
        change.notify_all();
    }
}

void PGArena::constructArena(const size_t k, const bool parallel, const bool print_product_states, const bool print_product_successors) {
    if (ptree) {
        buildProductAutomaton(k, parallel, print_product_states, print_product_successors);
        ptree.reset();
    }
}

void PGArena::print_basic_info() const {
    std::cout << "Number of env nodes  : " << n_env_nodes << std::endl;
    std::cout << "Number of env actions: " << n_env_actions << std::endl;
    std::cout << "Number of sys edges  : " << n_sys_edges << std::endl;
    std::cout << "Number of colors     : " << n_colors << std::endl;
}

void PGArena::print(std::ostream& out) const {
    switch(parity_type) {
        case Parity::ODD:
            std::cout << "min odd parity arena" << std::endl;
            break;
        case Parity::EVEN:
            std::cout << "min even parity arena" << std::endl;
            break;
        default:
            std::cout << "no parity arena" << std::endl;
    }
    size_t id_size = 1 + (size_t)log10(n_env_nodes * n_env_actions);
    size_t color_size = 1 + (size_t)log10(n_colors);

    out << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;
    out << "---- sys nodes -----" << std::endl;
    out << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;
    for (node_id_t i = 0; i < n_env_nodes * n_env_actions; ++i) {
        if (i == initial_node) {
            out << ">";
        }
        else {
            out << " ";
        }
        out << std::setw(id_size) << i;
        out << " [";
        for (edge_id_t j = sys_succs_begin[i]; j != sys_succs_begin[i+1]; ++j) {
            out << std::setw(id_size) << sys_succs[j] << ":" << std::setw(color_size) << colors[j];
            if (j + 1 != sys_succs_begin[i+1]) {
                out << ",";
            }
        }
        out << "]";
        out << std::endl;
    }
    out << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;
    out << "---- env nodes -----" << std::endl;
    out << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;
    for (node_id_t i = 0; i < n_env_nodes; ++i) {
        if (i == initial_node) {
            out << ">";
        }
        else {
            out << " ";
        }
        out << std::setw(id_size) << i;
        out << " [";
        const node_id_t succs_begin = i * n_env_actions;
        const node_id_t succs_end = succs_begin + n_env_actions;
        for (node_id_t j = succs_begin; j != succs_end; ++j) {
            out << std::setw(id_size) << j;
            if (j + 1 != succs_end) {
                out << ",";
            }
        }
        out << "]";
        out << std::endl;
    }
    out << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;
}

}
