#include "PGSISolver.h"

#include <iomanip>

namespace pg {

inline distance_t PGSISolver::color_distance_delta(const color_t& color) {
    return 1 - (((arena.parity_type + color) & 1) << 1);
}

PGSISolver::PGSISolver(pg::PGArena& arena, const bool parallel_construction, const int num_threads, const bool verbose) :
    PGSolver(arena, parallel_construction, num_threads, verbose)
{ }

PGSISolver::~PGSISolver() { }

void PGSISolver::strategy_iteration() {
    print_debug("Starting strategy iteration…");
    print_values_debug();

    bool change = true;
    while (change && !realizable) {
        bellman_ford();
        print_debug("Executing strategy improvement…");
        change = strategy_improvement();
        print_values_debug();
        print_debug("Marking solved nodes");
        update_nodes();
        print_values_debug();
    }
}

void PGSISolver::update_nodes() {
    dist_id_t k = 0;
    for (node_id_t i = 0; i < n_sys_nodes; i++, k += arena.n_colors) {
        if (sys_distances[k] == DISTANCE_INFINITY) {
            // node won by system player
            sys_distances[k] = DISTANCE_SOLVED;
            for (edge_id_t j = arena.sys_succs_begin[i]; j != arena.sys_succs_begin[i+1]; j++) {
                if (successors[j]) {
                    const node_id_t successor = arena.sys_succs[j];
                    if (env_distances[successor * arena.n_colors] < DISTANCE_INFINITY) {
                        successors[j] = false;
                    }
                }
            }
        }
    }
    k = 0;
    for (node_id_t i = 0; i < n_env_nodes; i++, k += arena.n_colors) {
        if (env_distances[k] == DISTANCE_INFINITY) {
            // node won by system player
            env_distances[k] = DISTANCE_SOLVED;
        }
    }
    realizable = (env_distances[arena.initial_node * arena.n_colors] == DISTANCE_SOLVED);
}

void PGSISolver::bellman_ford() {
    print_debug("Executing Bellman-Ford algorithm…");
    bellman_ford_init();
    print_values_debug();
    bool change = true;
    while (change) {
        print_debug("Executing Bellman-Ford sys iteration…");
        bellman_ford_sys_iteration();
        print_values_debug();
        print_debug("Executing Bellman-Ford env iteration…");
        change = bellman_ford_env_iteration();
        print_values_debug();
    }
}

void PGSISolver::bellman_ford_init() {
    #pragma omp parallel for
    for (node_id_t i = 0; i < n_env_nodes; i++) {
        const dist_id_t k = i * arena.n_colors;
        if (env_distances[k] != DISTANCE_SOLVED) {
            env_distances[k] = DISTANCE_INFINITY;
        }
    }
}

void PGSISolver::bellman_ford_sys_iteration() {
    #pragma omp parallel for
    for (node_id_t i = 0; i < n_sys_nodes; i++) {
        const dist_id_t k = i * arena.n_colors;
        if (sys_distances[k] != DISTANCE_SOLVED) {
            for (dist_id_t l = k; l < k + n_colors; l++) {
                sys_distances[l] = 0;
            }
            for (edge_id_t j = arena.sys_succs_begin[i]; j != arena.sys_succs_begin[i+1]; j++) {
                if (successors[j]) {
                    const node_id_t successor = arena.sys_succs[j];
                    dist_id_t m = successor * arena.n_colors;

                    if (env_distances[m] < DISTANCE_INFINITY) {
                        bool local_change = false;

                        const color_t cur_color = color_map[arena.colors[j]];
                        const distance_t cur_color_change = color_distance_delta(cur_color);
                        sys_distances[k + cur_color] -= cur_color_change;

                        for (dist_id_t l = k; l < k + n_colors; l++, m++) {
                            const distance_t d = sys_distances[l];
                            const distance_t d_succ = env_distances[m];
                            if (local_change || d_succ > d) {
                                sys_distances[l] = d_succ;
                                local_change = true;
                            }
                            else if (d_succ != d) {
                                break;
                            }
                        }

                        sys_distances[k + cur_color] += cur_color_change;
                    }
                    else {
                        sys_distances[k] = DISTANCE_INFINITY;
                        break;
                    }
                }
            }
        }
    }
}

bool PGSISolver::bellman_ford_env_iteration() {
    bool change = false;
    #pragma omp parallel for
    for (node_id_t i = 0; i < n_env_nodes; i++) {
        const dist_id_t k = i * arena.n_colors;
        const node_id_t succs_begin = i * arena.n_env_actions;
        if (env_distances[k] != DISTANCE_SOLVED) {
            for (edge_id_t successor = succs_begin; successor != succs_begin + arena.n_env_actions; successor++) {
                dist_id_t m = successor * arena.n_colors;

                if (sys_distances[m] < DISTANCE_INFINITY) {
                    bool local_change = false;

                    for (dist_id_t l = k; l < k + n_colors; l++, m++) {
                        const distance_t d = env_distances[l];
                        const distance_t d_succ = sys_distances[m];
                        if (local_change || d_succ < d) {
                            env_distances[l] = d_succ;
                            local_change = true;
                        }
                        else if (d_succ != d) {
                            break;
                        }
                    }
                    if (local_change) {
                        change = true;
                    }
                }
            }
        }
    }
    return change;
}

bool PGSISolver::strategy_improvement() {
    bool change = false;
    dist_id_t k = 0;
    for (node_id_t i = 0; i < n_sys_nodes; i++, k += arena.n_colors) {
        if (sys_distances[k] < DISTANCE_INFINITY) {
            for (edge_id_t j = arena.sys_succs_begin[i]; j != arena.sys_succs_begin[i+1]; j++) {
                successors[j] = false;
                const node_id_t successor = arena.sys_succs[j];

                if (successor < n_env_nodes) {
                    bool improvement = true;
                    dist_id_t m = successor * arena.n_colors;

                    const color_t cur_color = color_map[arena.colors[j]];
                    const distance_t cur_color_change = color_distance_delta(cur_color);
                    sys_distances[k + cur_color] -= cur_color_change;

                    for (dist_id_t l = k; l < k + n_colors; l++, m++) {
                        const distance_t d = sys_distances[l];
                        const distance_t d_succ = env_distances[m];
                        if (d_succ > d) {
                            // strict improvement
                            change = true;
                            break;
                        }
                        else if (d_succ != d) {
                            improvement = false;
                            break;
                        }
                    }

                    sys_distances[k + cur_color] += cur_color_change;

                    if (improvement) {
                        successors[j] = true;
                    }
                }
            }
        }
    }
    return change;
}

inline void PGSISolver::print_values_debug() {
    if (verbose) {
        std::cout << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;
        std::cout << "---- sys nodes -----" << std::endl;
        std::cout << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;
        std::cout << "        v     o(v)    d(v)" << std::endl;
        for (size_t i = 0; i < n_sys_nodes; ++i) {
            if (i == arena.initial_node) {
                std::cout << ">";
            }
            else {
                std::cout << " ";
            }
            std::cout << std::setw(8) << i;

            std::cout << " [";
            for (edge_id_t j = arena.sys_succs_begin[i]; j != arena.sys_succs_begin[i+1]; ++j) {
                if (successors[j]) {
                    std::cout << " " << arena.sys_succs[j];
                }
            }
            std::cout << " ]";

            for (color_t c = 0; c < n_colors; c++) {
                distance_t distance = sys_distances[i*arena.n_colors + c];
                if (distance == DISTANCE_INFINITY) {
                    std::cout << "       ∞";
                    break;
                }
                else if (distance == DISTANCE_SOLVED) {
                    std::cout << "  solved";
                    break;
                }
                else {
                    std::cout << std::setw(8) << distance;
                }
            }
            std::cout << std::endl;
        }
        std::cout << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;
        std::cout << "---- env nodes -----" << std::endl;
        std::cout << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;
        std::cout << "        v     d(v)" << std::endl;
        for (size_t i = 0; i < n_env_nodes; ++i) {
            if (i == arena.initial_node) {
                std::cout << ">";
            }
            else {
                std::cout << " ";
            }
            std::cout << std::setw(8) << i;
            for (color_t c = 0; c < n_colors; c++) {
                distance_t distance = env_distances[i*arena.n_colors + c];
                if (distance == DISTANCE_INFINITY) {
                    std::cout << "       ∞";
                    break;
                }
                else if (distance == DISTANCE_SOLVED) {
                    std::cout << "  solved";
                    break;
                }
                else {
                    std::cout << std::setw(8) << distance;
                }
            }
            std::cout << std::endl;
        }
        std::cout << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;
    }
}

void PGSISolver::reset() {
    realizable = false;

    env_distances.clear();
    sys_distances.clear();
    successors.clear();
}

void PGSISolver::solve_game() {
    env_distances.resize(n_env_nodes * arena.n_colors, 0);
    sys_distances.resize(n_sys_nodes * arena.n_colors, 0);
    successors.resize(n_sys_edges, false);

    strategy_iteration();
}

}
