#include "PGSolver.h"

#include <bitset>
#include <deque>
#include <set>
#include <mutex>

#include "omp.h"

namespace pg {

PGSolver::PGSolver(pg::PGArena& arena, const bool parallel_construction, const int num_threads, const bool verbose) :
    arena(arena),
    parallel_construction(parallel_construction),
    num_threads(num_threads),
    verbose(verbose),
    realizable(false),
    n_colors(arena.n_colors)
{
    init_solver();
}

PGSolver::~PGSolver() { }

void PGSolver::init_solver() {
    n_env_nodes = arena.n_env_nodes;
    n_sys_edges = arena.n_sys_edges;
    n_sys_nodes = n_env_nodes * arena.n_env_actions;
}

void PGSolver::compact_colors() {
    n_colors = arena.n_colors;

    std::vector<edge_id_t> color_count(n_colors, 0);

    for (edge_id_t i = 0; i < n_sys_edges; ++i) {
        const color_t c = arena.colors[i];
        color_count[c]++;
    }

    color_map.resize(n_colors);

    color_t cur_color = 0;
    for (color_t c = 0; c < n_colors; ++c) {
        if (color_count[c] != 0) {
            if ((c % 2) != (cur_color % 2)) {
                cur_color++;
            }
            color_map[c] = cur_color;
        }
    }

    n_colors = cur_color + 1;
}

void PGSolver::solve() {
    // set number of threads for parallel solving
    int max_threads;
    if (num_threads > 0) {
        max_threads = num_threads;
    }
    else {
        max_threads = omp_get_max_threads();
    }
    if (parallel_construction && max_threads > 1) {
        // leave one thread for construction of arena
        max_threads--;
    }
    omp_set_num_threads(max_threads);

    if (parallel_construction) {
        while (!arena.solved) {
            {
                std::unique_lock<std::mutex> lock(arena.size_mutex);
                while (!arena.complete && arena.n_env_nodes == n_env_nodes) {
                    arena.change.wait(lock);
                }

                init_solver();
            }

            std::unique_lock<std::mutex> lock(arena.arena_mutex);
            compact_colors();
            solve_game();

            if (realizable || (arena.complete && arena.n_env_nodes == n_env_nodes)) {
                arena.solved = true;
            }
        }
    }
    else {
        init_solver();
        compact_colors();
        solve_game();
    }
}

void PGSolver::print_debug(const std::string& str) const {
    if (verbose) {
        std::cout << str << std::endl;
    }
}

bool PGSolver::is_realizable() const {
    return realizable;
}

void PGSolver::construct_automaton(controller::MealyMachine& m) const {
    std::vector<node_id_t> state_map(n_env_nodes, n_env_nodes);

    std::deque<node_id_t> queue;
    queue.push_back(arena.initial_node);

    controller::state_id_t initial_state = m.addState();
    state_map[arena.initial_node] = initial_state;

    while (!queue.empty()) {
        node_id_t env_node = queue.front();
        queue.pop_front();

        controller::state_id_t state = state_map[env_node];

        node_id_t sys_node = env_node * arena.n_env_actions;
        for (ltl::letter_t input = 0; input < arena.n_env_actions; ++input, ++sys_node) {
            std::map<node_id_t, std::vector<ltl::letter_t>> successor_list;

            for (edge_id_t j = arena.sys_succs_begin[sys_node]; j != arena.sys_succs_begin[sys_node+1]; ++j) {
                if (successors[j]) {
                    node_id_t successor = arena.sys_succs[j];
                    ltl::letter_t output = arena.sys_output[j];

                    auto list = successor_list.find(successor);
                    if (list == successor_list.end()) {
                        successor_list.insert({ successor, { output } });
                    }
                    else {
                        list->second.push_back(output);
                    }
                }
            }

            // Heuristic: if possible, choose successor that was already explored,
            // and among those, choose successor with with maximum non-determinism

            auto succ_it = successor_list.begin();
            size_t succ_num_outputs = 0;
            bool succ_explored = false;

            for (auto it = successor_list.begin(); it != successor_list.end(); it++) {
                const node_id_t cur_successor = it->first;
                size_t cur_succ_num_outputs = it->second.size();
                bool cur_succ_explored = (state_map[cur_successor] != arena.n_env_nodes);
                if (
                        (cur_succ_explored && !succ_explored) ||
                        (cur_succ_explored == succ_explored && cur_succ_num_outputs > succ_num_outputs)
                ) {
                    succ_it = it;
                    succ_num_outputs = cur_succ_num_outputs;
                    succ_explored = cur_succ_explored;
                }
            }

            const node_id_t successor = succ_it->first;
            const std::vector<ltl::letter_t>& outputs = succ_it->second;

            if (state_map[successor] == n_env_nodes) {
                controller::state_id_t successor_state = m.addState();
                state_map[successor] = successor_state;
                queue.push_back(successor);
            }

            m.setSuccessor(state, input, state_map[successor], outputs);
        }
    }
}

}
