#include "ParityAutomatonTree.h"

namespace pg {

ParityAutomatonTree::ParityAutomatonTree(
        const NodeType node_type, const Parity parity_type, const color_t max_color
) :
    node_type(node_type),
    parity_type(parity_type),
    max_color(max_color)
{ }

ParityAutomatonTreeLeaf::ParityAutomatonTreeLeaf(
        const NodeType node_type, const Parity parity_type, const color_t max_color,
        const owl::Automaton _automaton
) :
    ParityAutomatonTree(node_type, parity_type, max_color),
    automaton(std::move(_automaton)),
    mapping(automaton.alphabet_mapping())
{ }

ParityAutomatonTreeNode::ParityAutomatonTreeNode(
        const NodeType node_type, const Parity parity_type, const color_t max_color, const color_t max_child_color,
        std::vector<std::unique_ptr<ParityAutomatonTree>> _children
) :
    ParityAutomatonTree(node_type, parity_type, max_color),
    children(std::move(_children)),
    max_child_color(max_child_color)
{ }

ParityAutomatonTree::~ParityAutomatonTree() { }

ParityAutomatonTreeLeaf::~ParityAutomatonTreeLeaf() { }

ParityAutomatonTreeNode::~ParityAutomatonTreeNode() { }

std::unique_ptr<ParityAutomatonTree> ParityAutomatonTree::constructTree(const owl::LabelledTree<owl::Tag, owl::Automaton>& tree) {
    switch (tree.type) {
        case owl::LEAF: {
            const owl::Automaton automaton = tree.getLabel2();
            const owl::Acceptance acceptance = automaton.acceptance();
            NodeType node_type;
            Parity parity_type;
            color_t max_color;
            switch (acceptance) {
                case owl::PARITY_MAX_EVEN:
                    throw std::invalid_argument("max parity not supported");
                case owl::PARITY_MAX_ODD:
                    throw std::invalid_argument("max parity not supported");
                case owl::PARITY_MIN_EVEN:
                    node_type = NodeType::PARITY;
                    parity_type = Parity::EVEN;
                    max_color = automaton.acceptance_set_count() - 1;
                    break;
                case owl::PARITY_MIN_ODD:
                    node_type = NodeType::PARITY;
                    parity_type = Parity::ODD;
                    max_color = automaton.acceptance_set_count() - 1;
                    break;
                case owl::BUCHI:
                    node_type = NodeType::BUCHI;
                    parity_type = Parity::EVEN;
                    max_color = 1;
                    break;
                case owl::CO_BUCHI:
                    node_type = NodeType::CO_BUCHI;
                    parity_type = Parity::ODD;
                    max_color = 1;
                    break;
                case owl::SAFETY:
                    node_type = NodeType::SAFETY;
                    parity_type = Parity::EVEN;
                    max_color = 1;
                    break;
                case owl::CO_SAFETY:
                    node_type = NodeType::CO_SAFETY;
                    parity_type = Parity::ODD;
                    max_color = 1;
                    break;
                default:
                    throw std::invalid_argument("unsupported acceptance");
            }
            std::unique_ptr<ParityAutomatonTree> leaf(new ParityAutomatonTreeLeaf(node_type, parity_type, max_color, std::move(automaton)));
            return leaf;
        }
        case owl::NODE: {
            const owl::Tag tag = tree.getLabel1();
            NodeType node_type;
            Parity parity_type;
            color_t max_color = 1;
            color_t max_child_color = 1;
            size_t round_robin_size = 0;
            switch (tag) {
                case owl::CONJUNCTION:
                    node_type = NodeType::CONJUNCTION;
                    parity_type = Parity::ODD;
                    break;
                case owl::DISJUNCTION:
                    node_type = NodeType::DISJUNCTION;
                    parity_type = Parity::EVEN;
                    break;
                default:
                    throw std::invalid_argument("unsupported node");
            }
            std::vector<std::unique_ptr<ParityAutomatonTree>> children;
            bool complex_child = false;
            for (const owl::LabelledTree<owl::Tag, owl::Automaton>& child : tree.getChildren()) {
                std::unique_ptr<ParityAutomatonTree> pchild = ParityAutomatonTree::constructTree(child);
                if (pchild->node_type == NodeType::CONJUNCTION || pchild->node_type == NodeType::DISJUNCTION || pchild->node_type == NodeType::PARITY) {
                    if (complex_child) {
                        throw std::invalid_argument("unsupported automaton tree");
                    }
                    else {
                        complex_child = true;
                    }
                    if (parity_type == pchild->parity_type) {
                        max_child_color = std::max(max_child_color, pchild->max_color);
                    }
                    else {
                        max_child_color = std::max(max_child_color, (color_t)(pchild->max_color + 1));
                    }
                }
                else if (
                    (node_type == NodeType::CONJUNCTION && pchild->node_type == NodeType::BUCHI) ||
                    (node_type == NodeType::DISJUNCTION && pchild->node_type == NodeType::CO_BUCHI)
                ) {
                    round_robin_size++;
                }
                if (parity_type == pchild->parity_type) {
                    max_color = std::max(max_color, pchild->max_color);
                }
                else {
                    max_color = std::max(max_color, (color_t)(pchild->max_color + 1));
                }
                children.push_back(std::move(pchild));
            }
            if (round_robin_size != 0 && max_color % 2 != 0) {
                max_color++;
            }
            std::unique_ptr<ParityAutomatonTree> node(new ParityAutomatonTreeNode(node_type, parity_type, max_color, max_child_color, std::move(children)));
            return node;
        }
    }
    throw std::invalid_argument("unsupported automaton tree");
}

color_t ParityAutomatonTreeLeaf::getSuccessor(state_iter& it, product_state_t& new_state, ltl::letter_t letter) {
    const local_state_t local_state = *it; ++it;
    if (local_state == NODE_BOTTOM) {
        //std::cout << "⊥";
        new_state.push_back(NODE_BOTTOM);
        return 1 - parity_type;
    }
    else if (local_state == NODE_TOP) {
        //std::cout << "⊤";
        new_state.push_back(NODE_TOP);
        return parity_type;
    }

    if (local_state >= successors.size()) {
        successors.resize(local_state + 1);
    }

    if (successors[local_state].empty()) {
        if (node_type == NodeType::PARITY || node_type == NodeType::BUCHI || node_type == NodeType::CO_BUCHI) {
            const std::vector<owl::Edge> edges = automaton.edges(local_state);
            for (const auto& edge : edges) {
                local_state_t new_state;
                color_t color;
                if (edge.successor >= 0) {
                    new_state = edge.successor;
                    if (node_type == NodeType::BUCHI || node_type == NodeType::CO_BUCHI) {
                        if (edge.colour == -1) {
                            color = 1;
                        }
                        else {
                            color = 0;
                        }
                    }
                    else if (edge.colour == -1) {
                        color = max_color;
                    }
                    else {
                        color = edge.colour;
                    }
                }
                else {
                    new_state = NODE_BOTTOM;
                    color = 1 - parity_type;
                }
                successors[local_state].push_back( Edge(new_state, color) );
            }
        }
        else if (node_type == NodeType::SAFETY || node_type == NodeType::CO_SAFETY) {
            const std::vector<int> local_successors = automaton.successors(local_state);
            for (const auto successor : local_successors) {
                local_state_t new_state;
                color_t color;
                if (successor >= 0) {
                    new_state = successor;
                    color = 0;
                    if (node_type == NodeType::CO_SAFETY && successor == 1) {
                        new_state = NODE_TOP;
                        color = 1;
                    }
                }
                else {
                    new_state = NODE_BOTTOM;
                    color = 1 - parity_type;
                }
                successors[local_state].push_back( Edge(new_state, color) );
            }
        }
        //std::cout << "x";
    }
    else {
        //std::cout << ".";
    }

    ltl::letter_t new_letter = 0;
    for (const auto& map : mapping) {
        new_letter |= (letter & (1 << map.first)) >> (map.first - map.second);
    }

    Edge edge = successors[local_state][new_letter];
    new_state.push_back(edge.successor);

    return edge.color;
}

color_t ParityAutomatonTreeNode::getSuccessor(state_iter& it, product_state_t& new_state, ltl::letter_t letter) {
    local_state_t round_robin_counter = *it; ++it;

    if (round_robin_counter == NODE_BOTTOM) {
        //std::cout << "⊥";
        new_state.push_back(NODE_BOTTOM);
        return 1 - parity_type;
    }
    else if (round_robin_counter == NODE_TOP) {
        //std::cout << "⊤";
        new_state.push_back(NODE_TOP);
        return parity_type;
    }
    if (node_type == NodeType::CONJUNCTION) {
        //std::cout << "∧";
    }
    else {
        //std::cout << "∨";
    }

    local_state_t min_parity = *it; ++it;

    local_state_t buchi_index = 0;
    size_t node_index = new_state.size();
    size_t active_children = 0;
    new_state.resize(node_index + 2);

    for (auto &child : children) {
        size_t child_state_index = new_state.size();
        color_t child_color = child->getSuccessor(it, new_state, letter);

        if (new_state[child_state_index] == NODE_BOTTOM) {
            if (node_type == NodeType::CONJUNCTION) {
                // discard children
                new_state.resize(node_index);
                new_state.push_back(NODE_BOTTOM);
                return 1 - parity_type;
            }
        }
        else if (new_state[child_state_index] == NODE_TOP) {
            if (node_type == NodeType::DISJUNCTION) {
                // discard children
                new_state.resize(node_index);
                new_state.push_back(NODE_TOP);
                return parity_type;
            }
        }
        else {
            active_children++;
        }

        if (
            child->node_type == NodeType::DISJUNCTION ||
            child->node_type == NodeType::CONJUNCTION ||
            child->node_type == NodeType::PARITY
        ) {
            if (parity_type == child->parity_type) {
                min_parity = std::min(min_parity, child_color);
            }
            else {
                min_parity = std::min(min_parity, (color_t)(child_color + 1));
            }
        }
        else if (
            child_color == 0 && (
                (node_type == NodeType::CONJUNCTION &&
                    (child->node_type == NodeType::CO_SAFETY || child->node_type == NodeType::CO_BUCHI)
                ) ||
                (node_type == NodeType::DISJUNCTION &&
                    (child->node_type == NodeType::SAFETY || child->node_type == NodeType::BUCHI)
                )
            )
        ) {
            min_parity = 0;
        }
        else if (
            (node_type == NodeType::CONJUNCTION && child->node_type == NodeType::BUCHI) ||
            (node_type == NodeType::DISJUNCTION && child->node_type == NodeType::CO_BUCHI)
        ) {
            if (child_color == 0 && round_robin_counter == buchi_index) {
                round_robin_counter++;
            }
            buchi_index++;
        }
    }

    if (active_children == 0) {
        // discard children
        if (node_type == NodeType::CONJUNCTION) {
            new_state.resize(node_index);
            new_state.push_back(NODE_TOP);
            return parity_type;
        }
        else {
            new_state.resize(node_index);
            new_state.push_back(NODE_BOTTOM);
            return 1 - parity_type;
        }
    }
    else if (min_parity == 0 || round_robin_counter == buchi_index) {
        // output min_parity and reset
        new_state[node_index] = 0;
        new_state[node_index + 1] = max_child_color;
        return min_parity;
    }
    else {
        // output neutral color, not accepting for conjunction and accepting for disjunction
        new_state[node_index] = round_robin_counter;
        new_state[node_index + 1] = min_parity;
        return max_color;
    }
}

void ParityAutomatonTreeLeaf::getInitialState(product_state_t& state) const {
    // initial state of automaton
    state.push_back(0);
}

void ParityAutomatonTreeNode::getInitialState(product_state_t& state) const {
    // round-robin counter
    state.push_back(0);
    // minimal parity seen
    state.push_back(max_child_color);
    for (auto &child : children) {
        child->getInitialState(state);
    }
}

}
