#pragma once

#include <vector>
#include <map>
#include <memory>
#include <limits>

#include "owl.h"

#include "ltl/TLSF.h"
#include "PGArenaDefinitions.h"

namespace pg {

enum class NodeType {
    CONJUNCTION, DISJUNCTION, PARITY, BUCHI, CO_BUCHI, SAFETY, CO_SAFETY
};


typedef uint32_t local_state_t;
static_assert(sizeof(local_state_t) == sizeof(color_t), "sizes of local state type and color type do not match");

constexpr local_state_t NODE_TOP = std::numeric_limits<local_state_t>::max() - 1;
constexpr local_state_t NODE_BOTTOM = std::numeric_limits<local_state_t>::max() - 2;

typedef std::vector<local_state_t> product_state_t;
typedef product_state_t::const_iterator state_iter;

struct Edge {
    const local_state_t successor;
    const color_t color;

    Edge(const local_state_t successor, const color_t color) : successor(successor), color(color) {};
};

class ParityAutomatonTree {
    protected:
        ParityAutomatonTree(
            const NodeType node_type, const Parity parity_type, const color_t max_color
        );

    public:
        virtual ~ParityAutomatonTree();

        const NodeType node_type;
        const Parity parity_type;
        const color_t max_color;

        virtual color_t getSuccessor(state_iter& it, product_state_t& new_state, ltl::letter_t letter) = 0;
        virtual void getInitialState(product_state_t& state) const = 0;

        static std::unique_ptr<ParityAutomatonTree> constructTree(const owl::LabelledTree<owl::Tag, owl::Automaton>& tree);
};

class ParityAutomatonTreeLeaf : public ParityAutomatonTree {
    friend class ParityAutomatonTree;

    private:
        const owl::Automaton automaton;
        const std::map<int, int> mapping;
        std::vector<std::vector<Edge>> successors;

    protected:
        ParityAutomatonTreeLeaf(
            const NodeType node_type, const Parity parity_type, const color_t max_color,
            const owl::Automaton automaton
        );

    public:
        virtual ~ParityAutomatonTreeLeaf();

        color_t getSuccessor(state_iter& it, product_state_t& new_state, ltl::letter_t letter);
        void getInitialState(product_state_t& state) const;
};

class ParityAutomatonTreeNode : public ParityAutomatonTree {
    friend class ParityAutomatonTree;

    private:
        const std::vector<std::unique_ptr<ParityAutomatonTree>> children;
        const color_t max_child_color;

    protected:
        ParityAutomatonTreeNode(
            const NodeType node_type, const Parity parity_type, const color_t max_color, const color_t max_child_color,
            std::vector<std::unique_ptr<ParityAutomatonTree>> children
        );

    public:
        virtual ~ParityAutomatonTreeNode();

        color_t getSuccessor(state_iter& it, product_state_t& new_state, ltl::letter_t letter);
        void getInitialState(product_state_t& state) const;
};

}
