#include "OptionParser.h"

#include <iostream>
#include <boost/algorithm/string/predicate.hpp>

namespace strix {

OptionParser::OptionParser() :
    help_options_desc_("Options"),
    hidden_options_desc_(),
    positional_options_desc_(),
    options_desc_()
{
    help_options_desc_.add_options()
        ("help,h", "display this help text and exit")
        ("input-file-format,i", po::value<std::string>(), "input file format: ltl or tlsf")
        ("output-file-format,f", po::value<std::string>(), "output file format: spec or kiss")
        ("output-file,o", po::value<std::string>(), "output file")
        ("compact-colors", "compact the colors by removing unused colors")
        ("monolithic,m", "use monolithic construction of a single automaton")
        ("work-group-size,w", po::value<size_t>(), "set the work group size (0=auto)")
        ("increment,k", po::value<size_t>(), "set the increment size")
        ("parallel,p", "construct and solve arena in parallel")
        ("threads", po::value<int>(), "set the number of threads (default 0=auto)")
        ("splitting", po::value<std::string>(), "splitting: never, auto, always (default auto)")
        ("print-arena", "print the arena")
        ("print-arena-info", "print basic arena information")
        ("print-device-info", "print device information")
        ("print-formula", "print parsed and simplified formulas")
        ("print-tree", "print automaton tree")
        ("print-product-states", "print states of product construction")
        ("print-product-successors", "print successors of product construction")
        ("timing,t", "measure and print timing information")
        ("no-solving", "do not solve the game")
        ("realizability", "only check realizability")
        ("validate-jni", "validate JNI interface")
        ("verbose-solver", "verbose output for solver")
    ;
    hidden_options_desc_.add_options()
        ("input-file", po::value<std::string>(), "input file")
    ;
    po::options_description options;
    options_desc_.add(help_options_desc_).add(hidden_options_desc_);
    positional_options_desc_.add("input-file", -1);
}

OptionParser::~OptionParser() { }

void OptionParser::print_options() const {
    std::cout << help_options_desc_;
}

Options OptionParser::parse_options(const int argc, const char* argv[]) const {
    po::variables_map vm;

    po::store(po::command_line_parser(argc, argv).
            options(options_desc_).
            positional(positional_options_desc_).
            run(), vm);
    po::notify(vm);

    Options options;

    options.help = vm.count("help") > 0;
    options.compact_colors = vm.count("no-compact-colors") > 0;
    options.monolithic = vm.count("monolithic") > 0;
    options.solve_game = vm.count("no-solving") == 0;
    options.parallel = vm.count("parallel") > 0;
    options.realizability = vm.count("realizability") > 0;
    options.validate_jni = vm.count("validate-jni") > 0;
    options.verbose_solver = vm.count("verbose-solver") > 0;
    options.print_arena = vm.count("print-arena") > 0;
    options.print_arena_info = vm.count("print-arena-info") > 0;
    options.print_device_info = vm.count("print-device-info") > 0;
    options.print_formula = vm.count("print-formula") > 0;
    options.print_tree = vm.count("print-tree") > 0;
    options.print_product_states = vm.count("print-product-states") > 0;
    options.print_product_successors = vm.count("print-product-successors") > 0;
    if (options.print_product_successors) {
        options.print_product_states = true;
    }
    options.timing = vm.count("timing") > 0;

    if (vm.count("splitting")) {
        std::string splitting = vm["splitting"].as<std::string>();
        boost::algorithm::to_lower(splitting);
        if (splitting == "never") {
            options.splitting = owl::NEVER;
        }
        else if (splitting == "auto") {
            options.splitting = owl::AUTO;
        }
        else if (splitting == "always") {
            options.splitting = owl::ALWAYS;
        }
        else {
            throw std::invalid_argument("Invalid splitting: " + vm["splitting"].as<std::string>());
        }
    } else {
        options.splitting = owl::AUTO;
    }

    if (vm.count("threads")) {
        options.threads = vm["threads"].as<int>();
        if (options.threads < 0) {
            throw std::invalid_argument("Invalid number of threads: " + options.threads);
        }
    } else {
        options.threads = 0;
    }

    if (vm.count("increment")) {
        options.increment = vm["increment"].as<size_t>();
    } else {
        options.increment = 0;
    }

    if (vm.count("input-file")) {
        options.input_file = vm["input-file"].as<std::string>();
    } else {
        options.input_file = "";
    }

    if (vm.count("input-file-format")) {
        std::string input_file_format = vm["input-file-format"].as<std::string>();
        boost::algorithm::to_lower(input_file_format);
        if (input_file_format == "ltl") {
            options.input_file_format = InputFileFormat::LTL;
        }
        else if (input_file_format == "tlsf") {
            options.input_file_format = InputFileFormat::TLSF;
        }
        else {
            throw std::invalid_argument("Invalid input file format selected: " + vm["input-file-format"].as<std::string>());
        }
    } else {
        if (boost::algorithm::ends_with(options.input_file, ".ltl")) {
            options.input_file_format = InputFileFormat::LTL;
        }
        else if (boost::algorithm::ends_with(options.input_file, ".tlsf")) {
            options.input_file_format = InputFileFormat::TLSF;
        }
        else if (!options.input_file.empty()) {
            throw std::invalid_argument("Unknown input file format");
        }
    }

    if (vm.count("output-file")) {
        options.output_file = vm["output-file"].as<std::string>();
    } else {
        options.output_file = "";
    }

    if (vm.count("output-file-format")) {
        std::string output_file_format = vm["output-file-format"].as<std::string>();
        boost::algorithm::to_lower(output_file_format);
        if (output_file_format == "spec") {
            options.output_file_format = OutputFileFormat::SPEC;
        }
        else if (output_file_format == "kiss") {
            options.output_file_format = OutputFileFormat::KISS;
        }
        else {
            throw std::invalid_argument("Invalid output file format selected: " + vm["output-file-format"].as<std::string>());
        }
    } else {
        if (boost::algorithm::ends_with(options.output_file, ".spec")) {
            options.output_file_format = OutputFileFormat::SPEC;
        }
        else if (boost::algorithm::ends_with(options.output_file, ".kiss")) {
            options.output_file_format = OutputFileFormat::KISS;
        }
        else {
            options.output_file_format = OutputFileFormat::SPEC;
        }
    }


    return options;
}

}
