#include <iostream>
#include <iomanip>
#include <fstream>
#include <vector>
#include <memory>
#include <thread>

#include <boost/optional.hpp>

#include "OptionParser.h"
#include "util/Timer.h"
#include "ltl/TLSF.h"
#include "ltl/LTLParser.h"
#include "ltl/TLSFParser.h"
#include "controller/MealyMachine.h"
#include "pg/PGArena.h"
#include "pg/PGSolver.h"
#include "pg/PGSISolver.h"

#include "owl.h"

void print_tree(const owl::LabelledTree<owl::Tag, owl::Automaton>& tree, int indent) {
    std::cout << std::setfill(' ') << std::setw(2*indent) << "";

    if (tree.type == owl::LEAF) {
        const owl::Automaton& automaton = tree.getLabel2();
        const owl::Acceptance acceptance = automaton.acceptance();
        std::cout << "* Automaton with acceptance: ";
        switch(acceptance) {
            case owl::PARITY_MIN_EVEN:
                std::cout << "parity min even";
                break;
            case owl::PARITY_MAX_EVEN:
                std::cout << "parity max even";
                break;
            case owl::PARITY_MIN_ODD:
                std::cout << "parity min odd";
                break;
            case owl::PARITY_MAX_ODD:
                std::cout << "parity max odd";
                break;
            case owl::BUCHI:
                std::cout << "buchi";
                break;
            case owl::CO_BUCHI:
                std::cout << "co-buchi";
                break;
            case owl::SAFETY:
                std::cout << "safety";
                break;
            case owl::CO_SAFETY:
                std::cout << "co-safety";
                break;
        }
        std::cout << " (" << automaton.acceptance_set_count() << ")";
        std::map<int, int> mapping = automaton.alphabet_mapping();
        std::cout << ", mapping:" << std::endl;
        for (const auto& entry : mapping) {
            std::cout << std::setfill(' ') << std::setw(2*(indent+2)) <<
                         entry.first << " -> " << entry.second << std::endl;
        }
    } else {
        if (tree.getLabel1() == owl::CONJUNCTION) {
            std::cout << "* Conjunction" << std::endl;
        } else {
            std::cout << "* Disjunction" << std::endl;
        }

        for (auto const& child : tree.getChildren()) {
            print_tree(child, indent + 1);
        }
    }
}

owl::LabelledTree<owl::Tag, owl::Automaton> constructAutomatonTree(owl::OwlThread& owl, ltl::TLSF& tlsf, const bool monolithic, const bool on_the_fly, const owl::SafetySplitting splitting) {
    if (monolithic) {
        owl::Automaton automaton = owl.createAutomaton(tlsf.formula, on_the_fly);
        return owl::LabelledTree<owl::Tag, owl::Automaton>(automaton);
    }
    else {
        return owl.createAutomatonTree(tlsf.formula, on_the_fly, splitting);
    }
}

int main(const int argc, const char* argv[]) {

    try {

        // Parse options
        strix::OptionParser option_parser;
        strix::Options options = option_parser.parse_options(argc, argv);

        if (options.help) {
            std::cout << "Usage: " << argv[0] << " [OPTIONS] FILE" << std::endl;
            std::cout << "Synthesize a controller according to the LTL specification in the given FILE." << std::endl;
            std::cout << std::endl;
            option_parser.print_options();
            return EXIT_SUCCESS;
        }

        // Start main part
        Timer total_timer(options.timing);
        Timer timer(options.timing);
        total_timer.start("checking realizability");

        timer.start("initializing OWL-JNI");
        const char* classpath = "-Djava.class.path=" OWL_CLASSPATH;

        owl::OwlJavaVM owlJavaVM(classpath, options.validate_jni);
        owl::OwlThread owl = owlJavaVM.attachCurrentThread();

        timer.stop();

        timer.start("parsing input file");
        std::unique_ptr<ltl::Parser> parser;
        if (options.input_file.empty()) {
            throw std::invalid_argument("No input file was given.");
        }
        if (options.input_file_format == strix::InputFileFormat::LTL) {
            parser = std::unique_ptr<ltl::Parser>(new ltl::LTLParser(owl));
        }
        else if (options.input_file_format == strix::InputFileFormat::TLSF) {
            parser = std::unique_ptr<ltl::Parser>(new ltl::TLSFParser(owl));
        }
        else {
            throw std::invalid_argument("Unsupported input format");
        }
        ltl::TLSF tlsf = parser->parse(options.input_file, options.print_formula);
        timer.stop();

        timer.start("constructing automaton tree");
        owl::LabelledTree<owl::Tag, owl::Automaton> tree = constructAutomatonTree(owl, tlsf, options.monolithic, options.parallel, options.splitting);
        if (options.print_tree) {
            print_tree(tree, 0);
        }
        timer.stop();

        if (options.parallel) {
            timer.start("initializing arena");
        }
        else {
            timer.start("constructing arena");
        }
        pg::PGArena arena(tlsf, tree);
        if (!options.parallel) {
            arena.constructArena(options.increment, false, options.print_product_states, options.print_product_successors);
        }
        timer.stop();

        timer.start("initializing solver");
        std::unique_ptr<pg::PGSolver> solver = std::unique_ptr<pg::PGSolver>(new pg::PGSISolver(arena, options.parallel, options.threads, options.verbose_solver));
        timer.stop();

        if (options.solve_game) {
            if (options.parallel) {
                timer.start("solving game in parallel with construction of arena");
                std::thread solver_thread = std::thread(&pg::PGSolver::solve, &*solver);
                arena.constructArena(options.increment, true, options.print_product_states, options.print_product_successors);
                solver_thread.join();
            }
            else {
                timer.start("solving game");
                solver->solve();
            }
            timer.stop();
        }

        if (options.print_arena_info) {
            timer.start("writing arena info");
            arena.print_basic_info();
            timer.stop();
        }

        std::streambuf* buf;
        std::ofstream of;
        if (options.output_file.empty()) {
            buf = std::cout.rdbuf();
            options.output_file = "stdout";
        }
        else {
            of.open(options.output_file, std::ios::out);
            if (!of) {
                throw std::invalid_argument("Could not open output file: " + options.output_file);
            }
            buf = of.rdbuf();
        }
        std::ostream out(buf);
        if (options.print_arena) {
            timer.start("writing arena");
            arena.print(out);
            timer.stop();
        }
        total_timer.stop();

        if (options.solve_game) {
            if (solver->is_realizable()) {
                std::cout << "REALIZABLE" << std::endl;
                if (!options.realizability) {
                    total_timer.start("constructing solution");
                    controller::MealyMachine m(tlsf);
                    solver->construct_automaton(m);
                    m.reduce();
                    if (options.output_file_format == strix::OutputFileFormat::SPEC) {
                        m.print_spec(out);
                    }
                    else if (options.output_file_format == strix::OutputFileFormat::KISS) {
                        m.print_kiss(out);
                    }
                    else {
                        throw std::invalid_argument("Unsupported output format");
                    }
                    total_timer.stop();
                }
            }
            else {
                std::cout << "UNREALIZABLE" << std::endl;
            }
        }
        of.close();
    }
    catch (const std::exception& e) {
        std::cerr << "Error: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (const std::string& e) {
        std::cerr << "Error: " << e << std::endl;
        return EXIT_FAILURE;
    }
    catch (...) {
        std::cerr << "Error: Unknown error." << std::endl;
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
